# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import pathlib
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple

import click
import click_log
import IPython
from click import Parameter, Path, argument, option
from traitlets.config import Config

from .analysis_output import AnalysisOutput
from .context import Context, pass_context
from .db import DB
from .extensions import prompt_extension
from .filesystem import find_root
from .models import PrimaryKeyGenerator
from .pipeline import Pipeline
from .pipeline.add_features import AddFeatures
from .pipeline.create_database import CreateDatabase
from .pipeline.database_saver import DatabaseSaver
from .pipeline.model_generator import ModelGenerator
from .pipeline.trim_trace_graph import TrimTraceGraph
from .ui import filters
from .ui.interactive import Interactive
from .ui.server import start_server
from .warning_messages import update_warning_messages

MARKER_DIRECTORIES = [".pyre", ".hg", ".git", ".svn"]

# pyre-fixme[5]: Global expression must be annotated.
logger = logging.getLogger("sapp")


def require_option(current_ctx: click.Context, param_name: str) -> None:
    """Throw an exception if an option wasn't required. This is useful when its
    optional in some contexts but required for a subcommand"""

    ctx = current_ctx
    param_definition = None
    while ctx is not None:
        # ctx.command.params has the actual definition of the param. We use
        # this when raising the exception.
        param_definition = next(
            (p for p in ctx.command.params if p.name == param_name), None
        )

        # ctx.params has the current value of the parameter, as set by the user.
        if ctx.params.get(param_name):
            return
        ctx = ctx.parent

    assert param_definition, f"unknown parameter {param_name}"
    raise click.MissingParameter(ctx=current_ctx, param=param_definition)


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def common_options(func):
    @click.group(context_settings={"help_option_names": ["--help", "-h"]})
    @click_log.simple_verbosity_option(logger)
    @option(
        "--repository",
        "-r",
        default=lambda: find_root(MARKER_DIRECTORIES),
        type=Path(exists=True, file_okay=False),
        help="Root of the repository (regardless of the directory analyzed)",
    )
    @option(
        "--database-name",
        "--dbname",
        callback=default_database,
        type=Path(dir_okay=False),
    )
    @wraps(func)
    # pyre-fixme[53]: Captured variable `func` is not annotated.
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def default_database(
    ctx: click.Context, _param: Parameter, value: Optional[str]
) -> str:
    """By default, use a database at the current dir"""
    if value:
        return value

    return os.path.join(os.path.curdir, DB.DEFAULT_DB_FILE)


@click.command(
    help="interactive exploration of issues",
    context_settings={"ignore_unknown_options": True},
)
@pass_context
@click.argument("ipython_args", nargs=-1, type=click.UNPROCESSED)
# pyre-fixme[2]: Parameter must be annotated.
def explore(ctx: Context, ipython_args) -> None:
    scope_vars = Interactive(
        database=ctx.database,
        repository_directory=ctx.repository,
        parser_class=ctx.parser_class,
    ).setup()
    config = Config()
    config.InteractiveShellApp.extensions = [
        prompt_extension.__name__
    ] + ctx.ipython_extensions
    config.InteractiveShellApp.profile = "sapp"
    config.InteractiveShellApp.display_banner = False

    config.InteractiveShell.show_rewritten_input = False
    config.InteractiveShell.autocall = 2

    IPython.start_ipython(
        argv=ipython_args if ipython_args else [], user_ns=scope_vars, config=config
    )


@click.command(help="parse static analysis output and save to disk")
@pass_context
@option("--run-kind", type=str)
@option("--branch", type=str)
@option("--commit-hash", type=str)
@option("--job-id", type=str)
@option("--differential-id", type=int)
@option("--add-feature", type=str, multiple=True)
@option(
    "--previous-issue-handles",
    type=Path(exists=True),
    help=("file containing list of issue handles to compare INPUT_FILE to"),
)
@option(
    "--linemap",
    type=Path(exists=True),
    help="json file mapping new locations to old locations",
)
@option(
    "--store-unused-models",
    is_flag=True,
    help="store pre/post conditions unrelated to an issue",
)
@option("--dry-run", is_flag=True)
@argument("input_file", type=Path(exists=True))
def analyze(
    ctx: Context,
    run_kind: Optional[str],
    branch: Optional[str],
    commit_hash: Optional[str],
    job_id: Optional[str],
    differential_id: Optional[int],
    previous_issue_handles: Optional[str],
    linemap: Optional[str],
    store_unused_models: bool,
    dry_run: bool,
    input_file: str,
    add_feature: Optional[List[str]],
) -> None:
    # Store all options in the right places
    summary_blob: Dict[str, Any] = {
        "run_kind": run_kind,
        "repository": ctx.repository,
        "branch": branch,
        "commit_hash": commit_hash,
        "old_linemap_file": linemap,
        "store_unused_models": store_unused_models,
    }

    if job_id is None and differential_id is not None:
        job_id = "user_input_" + str(differential_id)
    summary_blob["job_id"] = job_id

    if previous_issue_handles:
        summary_blob["previous_issue_handles"] = AnalysisOutput.from_file(
            previous_issue_handles
        )

    # Construct pipeline
    if os.path.isdir(input_file):
        analysis_output = AnalysisOutput.from_directory(input_file)
    else:
        analysis_output = AnalysisOutput.from_file(input_file)

    pipeline_steps = [
        ctx.parser_class(),
        CreateDatabase(ctx.database),
        AddFeatures(add_feature),
        ModelGenerator(),
        TrimTraceGraph(),
        DatabaseSaver(ctx.database, PrimaryKeyGenerator(), dry_run),
    ]
    # pyre-fixme[6]: Expected
    #  `List[tools.sapp.sapp.pipeline.PipelineStep[typing.Any, typing.Any]]` for 1st
    #  param but got `List[typing.Union[DatabaseSaver, ModelGenerator, TrimTraceGraph,
    #  tools.sapp.sapp.base_parser.BaseParser]]`.
    pipeline = Pipeline(pipeline_steps)
    pipeline.run(analysis_output, summary_blob)


@click.command(
    help="backend flask server for exploration of issues",
    context_settings={"ignore_unknown_options": True},
)
@option("--debug/--no-debug", default=False, help="Start Flask server in debug mode")
@option(
    "--static-resources", default=None, help="Directory to serve static resources from"
)
# pyre-fixme[56]: Pyre was not able to infer the type of argument `os.getcwd()` to
#  decorator factory `click.option`.
@option(
    "--source-directory", default=os.getcwd(), help="Directory to look for source code"
)
@option(
    "--editor-schema",
    default=None,
    help="Editor schema to open files from a browser, e.g. `vscode:`",
)
@pass_context
def server(
    ctx: Context,
    debug: bool,
    static_resources: Optional[str],
    source_directory: str,
    editor_schema: Optional[str],
) -> None:
    start_server(ctx.database, debug, static_resources, source_directory, editor_schema)


@click.group()
def filter() -> None:
    pass


@filter.command(
    name="import",
    help="Import a filter or a directory containing filters into database",
)
@pass_context
@argument("input_filter_path", type=Path(exists=True, readable=True))
def import_filters(
    ctx: Context,
    input_filter_path: str,
) -> None:
    filters.import_filter_from_path(ctx.database, pathlib.Path(input_filter_path))


@filter.command(
    name="export",
    help="Export a filter from database",
)
@pass_context
@argument("filter_name", type=str)
@option(
    "--output-path",
    "-o",
    type=Path(writable=True),
    help="Path where you want to save the exported filter to",
)
def export_filter(
    ctx: Context,
    filter_name: str,
    output_path: str,
) -> None:
    if output_path:
        filters.export_filter(ctx.database, filter_name, pathlib.Path(output_path))
    else:
        filters.export_filter(ctx.database, filter_name)


@filter.command(name="delete", help="Delete filters from database")
@argument("filter_names", nargs=-1, required=True)
@pass_context
def delete_filters(
    ctx: Context,
    filter_names: Tuple[str],
) -> None:
    filters.delete_filters(ctx.database, filter_names)


@filter.command(
    name="issues",
    help="Show issues after applying a filter or directory of filters to a run",
    context_settings={"token_normalize_func": str.lower},
)
@click.option(
    "--output-format",
    type=click.Choice(["sapp", "sarif"]),
    default="sapp",
    help="output format you want your filtered results in",
)
@argument("run_id", type=int)
@argument("input_filter_path", type=Path(exists=True, readable=True))
@pass_context
def filter_issues(
    ctx: Context,
    run_id: int,
    input_filter_path: str,
    output_format: str,
) -> None:
    """Applies filter from INPUT_FILTER_PATH to filter issues in RUN_ID

    RUN_ID is the `Run` number corresponding to the list of issues you want to apply the filter to
    INPUT_FILTER_PATH is the path to the filter you want to use to filter the list of issues in RUN_ID
    OUTPUT_FORMAT is the format you want the results to be in
    """
    filters.filter_run(ctx, run_id, pathlib.Path(input_filter_path), output_format)


@click.group()
def update() -> None:
    pass


@update.command(
    name="warning-codes",
    help="Parse static analysis metadata output and update warning codes in SAPP db",
)
@argument("input_metadata_file", type=Path(exists=True, readable=True))
@pass_context
def update_warning_codes(
    ctx: Context,
    input_metadata_file: str,
) -> None:
    update_warning_messages(ctx.database, pathlib.Path(input_metadata_file))


commands: List[Callable[[], None]] = [
    analyze,
    explore,
    server,
    filter,
    update,
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import datetime
import logging
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import ujson as json

from ..models import (
    DBID,
    FrameReachability,
    Issue,
    IssueDBID,
    IssueInstance,
    IssueInstanceFixInfo,
    IssueStatus,
    PurgeStatus,
    Run,
    RunStatus,
    SharedText,
    SharedTextKind,
    SourceLocation,
    TraceFrame,
    TraceFrameAnnotation,
    TraceKind,
)
from ..trace_graph import LeafMapping, TraceGraph
from . import (
    DictEntries,
    ParseConditionTuple,
    ParseFeature,
    ParseIssueConditionTuple,
    ParseIssueTuple,
    ParseLeaf,
    ParseTraceAnnotation,
    ParseTypeInterval,
    PipelineStep,
    Summary,
)

log: logging.Logger = logging.getLogger("sapp")


# pyre-fixme[13]: Attribute `graph` is never initialized.
# pyre-fixme[13]: Attribute `summary` is never initialized.
class ModelGenerator(PipelineStep[DictEntries, TraceGraph]):
    def __init__(self) -> None:
        super().__init__()
        self.summary: Summary
        self.graph: TraceGraph
        self.visited_frames: Dict[int, Set[int]] = {}  # frame id -> leaf ids

    def run(self, input: DictEntries, summary: Summary) -> Tuple[TraceGraph, Summary]:
        self.summary = summary

        self.summary["trace_entries"] = defaultdict(
            lambda: defaultdict(list)
        )  # : Dict[TraceKind, Dict[Tuple[str, str], List[ParseCondition]]]
        self.summary["missing_traces"] = defaultdict(
            set
        )  # Dict[TraceKind, Set[Tuple[str, str]]]
        self.summary["big_tito"] = set()  # Set[Tuple[str, str, int]]

        self.graph = TraceGraph()
        self.summary["run"] = self._create_empty_run(status=RunStatus.INCOMPLETE)
        self.summary["run"].id = DBID()

        self.summary["trace_entries"][TraceKind.precondition] = input["preconditions"]
        self.summary["trace_entries"][TraceKind.postcondition] = input["postconditions"]
        callables = self._compute_callables_count(input["issues"])

        log.info("Generating issues and traces")
        for entry in input["issues"]:
            self._generate_issue(self.summary["run"], entry, callables)

        if self.summary.get("store_unused_models"):
            for trace_kind, traces in self.summary["trace_entries"].items():
                for _key, entry in traces:
                    self._generate_trace_frame(trace_kind, self.summary["run"], entry)

        return self.graph, self.summary

    def _compute_callables_count(
        self, issues: Iterable[ParseIssueTuple]
    ) -> Dict[str, int]:
        """Iterate over all issues and count the number of times each callable
        is seen."""
        count: Dict[str, int] = dict.fromkeys(
            [issue.callable for issue in issues], int(0)
        )
        for issue in issues:
            count[issue.callable] += 1

        return count

    def _create_empty_run(
        self,
        status: str = RunStatus.FINISHED,
        status_description: Optional[str] = None,
    ) -> Run:
        """setting boilerplate when creating a Run object"""
        run = Run(
            job_id=self.summary["job_id"],
            issue_instances=[],
            date=datetime.datetime.now(),
            status=status,
            status_description=status_description,
            repository=self.summary["repository"],
            branch=self.summary["branch"],
            commit_hash=self.summary["commit_hash"],
            kind=self.summary["run_kind"],
            purge_status=PurgeStatus.UNPURGED,
        )
        return run

    def _get_minimum_trace_length(
        self, entries: Iterable[ParseIssueConditionTuple]
    ) -> int:
        length = None
        for entry in entries:
            for (_leaf, depth) in entry.leaves:
                if length is None or length > depth:
                    length = depth
        if length is not None:
            return length
        return 0

    def _generate_issue(
        self, run: Run, entry: ParseIssueTuple, callablesCount: Dict[str, int]
    ) -> IssueInstance:
        """Insert the issue instance into a run. This includes creating (for
        new issues) or finding (for existing issues) Issue objects to associate
        with the instances.
        Also create sink entries and associate related issues"""

        trace_frames = []
        final_sink_kinds = set()
        initial_source_kinds = set()
        for p in entry.preconditions:
            tf, new_sink_ids = self._generate_issue_traces(
                TraceKind.PRECONDITION, run, entry, p
            )
            final_sink_kinds.update(new_sink_ids)
            trace_frames.append(tf)

        for p in entry.postconditions:
            tf, new_source_ids = self._generate_issue_traces(
                TraceKind.POSTCONDITION, run, entry, p
            )
            initial_source_kinds.update(new_source_ids)
            trace_frames.append(tf)

        callable = entry.callable

        source_details = {
            self._get_shared_text(SharedTextKind.SOURCE_DETAIL, name)
            for (name, _kind, _depth) in entry.initial_sources
            if name
        }
        sink_details = {
            self._get_shared_text(SharedTextKind.SINK_DETAIL, name)
            for (name, _kind, _depth) in entry.final_sinks
            if name
        }

        # pyre-fixme [9] Incompatible variable type: issue is declared to have type `Issue` but is used as type `munch.Munch`
        issue: Issue = Issue.Record(
            id=IssueDBID(),
            code=entry.code,
            handle=entry.handle,
            status=IssueStatus.UNCATEGORIZED,
            first_seen=run.date,
            run_id=run.id,
        )

        self.graph.add_issue(issue)

        fix_info = None
        fix_info_id = None
        if entry.fix_info is not None:
            fix_info = IssueInstanceFixInfo.Record(
                id=DBID(), fix_info=json.dumps(entry.fix_info)
            )
            fix_info_id = fix_info.id

        message = self._get_shared_text(SharedTextKind.MESSAGE, entry.message)
        filename_record = self._get_shared_text(SharedTextKind.FILENAME, entry.filename)
        callable_record = self._get_shared_text(SharedTextKind.CALLABLE, callable)

        # pyre-fixme [9] Incompatible variable type: issue is declared to have type `Issue` but is used as type `munch.Munch`
        instance: IssueInstance = IssueInstance.Record(
            id=DBID(),
            issue_id=issue.id,
            location=self.get_location(entry),
            filename_id=filename_record.id,
            callable_id=callable_record.id,
            run_id=run.id,
            fix_info_id=fix_info_id,
            message_id=message.id,
            rank=0,
            min_trace_length_to_sources=self._get_minimum_trace_length(
                entry.postconditions
            ),
            min_trace_length_to_sinks=self._get_minimum_trace_length(
                entry.preconditions
            ),
            callable_count=callablesCount[callable],
        )

        for sink in final_sink_kinds:
            self.graph.add_issue_instance_shared_text_assoc_id(instance, sink)
        for detail in sink_details:
            self.graph.add_issue_instance_shared_text_assoc(instance, detail)
        for source in initial_source_kinds:
            self.graph.add_issue_instance_shared_text_assoc_id(instance, source)
        for detail in source_details:
            self.graph.add_issue_instance_shared_text_assoc(instance, detail)

        if fix_info is not None:
            self.graph.add_issue_instance_fix_info(instance, fix_info)

        for trace_frame in trace_frames:
            self.graph.add_issue_instance_trace_frame_assoc(instance, trace_frame)

        for feature in entry.features:
            feature = self._get_shared_text(SharedTextKind.FEATURE, feature)
            self.graph.add_issue_instance_shared_text_assoc(instance, feature)

        self.graph.add_issue_instance(instance)
        return instance

    # We need to thread filename explicitly since the entry might be a callinfo.
    def _generate_tito(
        self,
        filename: str,
        entry: Union[
            ParseConditionTuple, ParseIssueConditionTuple, ParseTraceAnnotation
        ],
        callable: str,
    ) -> List[SourceLocation]:
        titos = list(entry.titos)
        if len(titos) > 200:
            pre_key: Tuple[str, str, int] = (filename, callable, len(titos))
            if pre_key not in self.summary["big_tito"]:
                log.info("Big Tito: %s", str(pre_key))
                self.summary["big_tito"].add(pre_key)
            titos = titos[:200]
        return titos

    def _generate_issue_traces(
        self,
        kind: TraceKind,
        run: Run,
        issue: ParseIssueTuple,
        callinfo: ParseIssueConditionTuple,
    ) -> Tuple[TraceFrame, Set[int]]:
        # Generates a synthetic trace frame from a forward or backward trace in callinfo
        # that represents a call edge from the issue callable to the start of a
        # a trace.
        # Generate all dependencies of this frame as well.
        caller = issue.callable
        titos = self._generate_tito(issue.filename, callinfo, caller)
        call_tf = self._generate_raw_trace_frame(
            kind,
            run=run,
            filename=issue.filename,
            caller=caller,
            caller_port="root",
            callee=callinfo.callee,
            callee_port=callinfo.port,
            callee_location=callinfo.location,
            leaves=callinfo.leaves,
            type_interval=callinfo.type_interval,
            titos=titos,
            annotations=callinfo.annotations,
            features=callinfo.features,
        )
        caller_leaf_ids = set()
        callee_leaf_ids = set()
        for leaf_map in call_tf.leaf_mapping:
            caller_leaf_ids.add(leaf_map.caller_leaf)
            callee_leaf_ids.add(leaf_map.callee_leaf)
        self._generate_transitive_trace_frames(run, call_tf, callee_leaf_ids)
        return call_tf, caller_leaf_ids

    def _generate_transitive_trace_frames(
        self, run: Run, start_frame: TraceFrame, outgoing_leaf_ids: Set[int]
    ) -> List[TraceFrame]:
        """Generates all trace frames reachable from start_frame, provided they contain
        a leaf_id from the initial set of leaf_ids. Also applies tito transforms
        in reverse, meaning it strips off local transforms from leaf kinds when
        necessary.

        Returns the TraceFrames associated this starting frame (generated or found existing)
        """
        returned_frames = []

        kind = start_frame.kind
        queue = [(start_frame, outgoing_leaf_ids)]
        while len(queue) > 0:
            frame, outgoing_leaves = queue.pop()
            if len(outgoing_leaves) == 0:
                continue

            returned_frames.append(frame)

            frame_id = frame.id.local_id
            if frame_id in self.visited_frames:
                outgoing_leaves = outgoing_leaves - self.visited_frames[frame_id]
                if len(outgoing_leaves) == 0:
                    continue
                else:
                    self.visited_frames[frame_id].update(outgoing_leaves)
            else:
                self.visited_frames[frame_id] = outgoing_leaves

            next_frames = self._get_or_populate_trace_frames(
                # pyre-fixme[6]: Expected `TraceKind` for 1st param but got `str`.
                kind,
                run,
                frame.callee_id,
                caller_port=frame.callee_port,
            )
            queue.extend(
                [
                    (
                        frame,
                        self.graph.compute_next_leaf_kinds(
                            outgoing_leaves, frame.leaf_mapping
                        ),
                    )
                    for frame in next_frames
                ]
            )
        return returned_frames

    def _get_or_populate_trace_frames(
        self, kind: TraceKind, run: Run, caller_id: DBID, caller_port: str
    ) -> List[TraceFrame]:
        if self.graph.has_trace_frames_with_caller(kind, caller_id, caller_port):
            return self.graph.get_trace_frames_from_caller(kind, caller_id, caller_port)
        key = (self.graph.get_text(caller_id), caller_port)
        new = [
            self._generate_trace_frame(kind, run, e)
            for e in self.summary["trace_entries"][kind].pop(key, [])
        ]
        if len(new) == 0 and not self.graph.is_leaf_port(key[1]):
            self.summary["missing_traces"][kind].add(key)
        return new

    def _generate_trace_frame(
        self, kind: TraceKind, run: Run, entry: ParseConditionTuple
    ) -> TraceFrame:
        titos = self._generate_tito(entry.filename, entry, entry.caller)
        return self._generate_raw_trace_frame(
            kind,
            run=run,
            filename=entry.filename,
            caller=entry.caller,
            caller_port=entry.caller_port,
            callee=entry.callee,
            callee_port=entry.callee_port,
            callee_location=entry.callee_location,
            titos=titos,
            leaves=entry.leaves,
            type_interval=entry.type_interval,
            annotations=entry.annotations,
            features=entry.features,
        )

    def _generate_raw_trace_frame(
        self,
        kind: TraceKind,
        run: Run,
        filename: str,
        caller: str,
        caller_port: str,
        callee: str,
        callee_port: str,
        callee_location: SourceLocation,
        titos: List[SourceLocation],
        leaves: Iterable[ParseLeaf],
        type_interval: Optional[ParseTypeInterval],
        annotations: Iterable[ParseTraceAnnotation],
        features: List[str],
    ) -> TraceFrame:
        leaf_kind = (
            SharedTextKind.SOURCE
            if kind is TraceKind.POSTCONDITION
            else SharedTextKind.SINK
        )
        lb, ub, preserves_type_context = self._get_interval(type_interval)
        caller_record = self._get_shared_text(SharedTextKind.CALLABLE, caller)
        callee_record = self._get_shared_text(SharedTextKind.CALLABLE, callee)
        filename_record = self._get_shared_text(SharedTextKind.FILENAME, filename)

        leaf_records = []
        leaf_mapping_ids: Set[LeafMapping] = set()
        for (leaf, depth) in leaves:
            leaf_record = self._get_shared_text(leaf_kind, leaf)
            caller_leaf_id = self.graph.get_transform_normalized_kind_id(leaf_record)
            callee_leaf_id = self.graph.get_transformed_kind_id(leaf_record)
            leaf_mapping_ids.add(
                LeafMapping(
                    caller_leaf=caller_leaf_id,
                    callee_leaf=callee_leaf_id,
                    transform=leaf_record.id.local_id,
                )
            )
            leaf_records.append((leaf_record, depth))

        trace_frame: TraceFrame = TraceFrame.Record(
            extra_fields=["leaf_mapping"],
            id=DBID(),
            kind=kind,
            caller_id=caller_record.id,
            caller_port=caller_port,
            callee_id=callee_record.id,
            callee_port=callee_port,
            callee_location=callee_location,
            filename_id=filename_record.id,
            titos=titos,
            run_id=run.id,
            preserves_type_context=preserves_type_context,
            type_interval_lower=lb,
            type_interval_upper=ub,
            leaf_mapping=leaf_mapping_ids,
            reachability=FrameReachability.UNREACHABLE,
        )

        for (leaf_record, depth) in leaf_records:
            self.graph.add_trace_frame_leaf_assoc(trace_frame, leaf_record, depth)

        # Note that the "graph._trace_frame_leaf_assoc" table is really associated with
        # the xdb table "trace_frame_message_assoc"
        # Putting the features into this assoc table is the correct thing to do, even though
        # the function to put it there is a bit odd.
        # Note that "graph._save_trace_frame_leaf_assoc" which is called by the bulk_saver
        # using "bulk_saver.add_trace_frame_leaf_assoc()" to drop into this table
        # as documented in models.py "class TraceFrameLeafAssoc(Base, PrepareMixin, RecordMixin)"
        for f in features:
            feature_record = self._get_shared_text(SharedTextKind.FEATURE, f)
            self.graph.add_trace_frame_leaf_assoc(trace_frame, feature_record, 0)

        self.graph.add_trace_frame(trace_frame)
        self._generate_trace_annotations(
            trace_frame.id, filename, caller, annotations, run
        )
        return trace_frame

    def _generate_issue_feature_contents(self, feature: ParseFeature) -> Set[str]:
        # Generates a synthetic feature from the extra/feature
        features = set()
        for key in feature:
            value = feature[key]
            if isinstance(value, str) and value:
                features.add(key + ":" + value)
            else:
                features.add(key)
        return features

    def _get_interval(
        self, ti: Optional[ParseTypeInterval]
    ) -> Tuple[Optional[int], Optional[int], bool]:
        if ti:
            return (ti.start, ti.finish, ti.preserves_type_context)
        else:
            return (None, None, False)

    def _generate_trace_annotations(
        self,
        parent_id: DBID,
        parent_filename: str,
        parent_caller: str,
        annotations: Iterable[ParseTraceAnnotation],
        run: Run,
    ) -> None:
        for annotation in annotations:
            location = annotation.location
            leaf_kind = annotation.leaf_kind
            kind = annotation.kind
            (trace_leaf_kind, trace_kind) = (
                (SharedTextKind.SINK, TraceKind.PRECONDITION)
                if kind == "tito_transform" or kind == "sink"
                else (SharedTextKind.SOURCE, TraceKind.POSTCONDITION)
            )
            annotation_record = TraceFrameAnnotation.Record(
                id=DBID(),
                trace_frame_id=parent_id,
                location=location,
                kind=kind,
                message=annotation.msg,
                leaf_id=(
                    None
                    if not leaf_kind
                    else self._get_shared_text(trace_leaf_kind, leaf_kind).id
                ),
                link=annotation.link,
                trace_key=annotation.trace_key,
            )
            self.graph.add_trace_annotation(annotation_record)

            for trace in annotation.subtraces:
                tf = self._generate_annotation_trace(
                    trace_kind, run, parent_filename, parent_caller, trace, annotation
                )
                self.graph.add_trace_frame_annotation_trace_frame_assoc(
                    annotation_record, tf
                )

    def _generate_annotation_trace(
        self,
        trace_kind: TraceKind,
        run: Run,
        parent_filename: str,
        parent_caller: str,
        trace: Dict[str, Any],
        annotation: ParseTraceAnnotation,
    ) -> TraceFrame:
        # Generates the first-hop trace frames from the annotation and
        # all dependencies of these sub traces. If this gets called, it is
        # assumed that the annotation leads to traces, and that the leaf kind
        # and depth are specified.
        callee = trace["callee"]
        callee_port = trace["port"]
        titos = self._generate_tito(parent_filename, annotation, parent_caller)
        call_tf = self._generate_raw_trace_frame(
            trace_kind,
            run,
            parent_filename,
            parent_caller,
            "root",
            callee,
            callee_port,
            annotation.location,
            titos,
            [(annotation.leaf_kind or "", annotation.leaf_depth)],
            annotation.type_interval,
            [],  # no more annotations for a precond coming from an annotation
            [],  # no breadcrumbs / features associated with the traces for annotations
        )
        self._generate_transitive_trace_frames(
            run, call_tf, {leaf_map.callee_leaf for leaf_map in call_tf.leaf_mapping}
        )
        return call_tf

    def _get_shared_text(self, kind: SharedTextKind, name: str) -> SharedText:
        return self.graph.get_or_add_shared_text(kind, name)

    @staticmethod
    def get_location(
        entry: ParseIssueTuple, is_relative: bool = False
    ) -> SourceLocation:
        line = entry.line
        if is_relative and entry.callable_line:
            line -= entry.callable_line
        return SourceLocation(line, entry.start, entry.end)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
from collections import defaultdict
from typing import DefaultDict, Dict, Iterable, List, Optional, Set, Tuple

from .bulk_saver import BulkSaver
from .models import (
    DBID,
    SHARED_TEXT_LENGTH,
    LeafMapping,
    Issue,
    IssueInstance,
    IssueInstanceFixInfo,
    SharedText,
    SharedTextKind,
    TraceFrame,
    TraceFrameAnnotation,
    TraceKind,
)

log: logging.Logger = logging.getLogger("sapp")

LeafIDToDepthMap = Dict[int, Optional[int]]


class TraceGraph(object):
    """Represents a graph of the Zoncolan trace steps. Nodes of the graph are
    the issues, preconditions, postconditions, sources and sinks. Edges are
    the the assocs, and for pre/postconditions, the map of 'caller->callee'
    gives a one direction edge to the next pre/postcondition, and the map of
    'callee->caller' gives the reverse edge.
    """

    def __init__(self) -> None:
        self._issues: Dict[int, Issue] = {}
        self._issue_instances: Dict[int, IssueInstance] = {}
        self._trace_annotations: Dict[int, TraceFrameAnnotation] = {}

        # Create a mapping of (caller_id, caller_port) to the corresponding
        # trace frame's id.
        self._trace_frames_map: DefaultDict[
            TraceKind, DefaultDict[Tuple[int, str], Set[int]]
        ] = defaultdict(lambda: defaultdict(set))

        # Similar to _trace_frames_map, but maps the reverse direction
        # of the trace graph, i.e. (callee_id, callee_port) to the
        # trace_frame_id.
        self._trace_frames_rev_map: DefaultDict[
            TraceKind, DefaultDict[Tuple[int, str], Set[int]]
        ] = defaultdict(lambda: defaultdict(set))

        self._trace_frames: Dict[int, TraceFrame] = {}

        self._shared_texts: Dict[int, SharedText] = {}
        self._shared_text_lookup: (
            DefaultDict[SharedTextKind, Dict[str, int]]
        ) = defaultdict(dict)

        self._trace_frame_leaf_assoc: DefaultDict[int, LeafIDToDepthMap] = defaultdict(
            lambda: {}
        )

        self._trace_frame_issue_instance_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)
        self._issue_instance_trace_frame_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)

        self._trace_frame_annotation_trace_frame_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)
        self._trace_frame_trace_frame_annotation_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)

        self._issue_instance_shared_text_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)
        self._shared_text_issue_instance_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)

        self._issue_instance_fix_info: Dict[int, IssueInstanceFixInfo] = {}

        # !!!!! IMPORTANT !!!!!
        # IF YOU ARE ADDING MORE FIELDS/EDGES TO THIS GRAPH, CHECK IF
        # TrimmedTraceGraph NEEDS TO BE UPDATED AS WELL.
        #
        # TrimmedTraceGraph will populate itself from this object. It searches
        # TrimmedGraph for nodes and edges of 'affected_files' and copies them
        # over. If new fields/edges are added, these may need to be copied in
        # TrimmedTraceGraph as well.

    def add_issue(self, issue: Issue) -> None:
        assert issue.id.local_id not in self._issues, "Issue already exists"
        self._issues[issue.id.local_id] = issue

    def get_issue(self, issue_id: DBID) -> Issue:
        return self._issues[issue_id.local_id]

    def get_issues(self) -> Iterable[Issue]:
        return (issue for issue in self._issues.values())

    def get_number_issues(self) -> int:
        return len(self._issues)

    def add_issue_instance(self, instance: IssueInstance) -> None:
        assert (
            instance.id.local_id not in self._issue_instances
        ), "Instance already exists"
        self._issue_instances[instance.id.local_id] = instance

    def get_issue_instances(self) -> Iterable[IssueInstance]:
        return (instance for instance in self._issue_instances.values())

    def add_issue_instance_fix_info(
        self, instance: IssueInstance, fix_info: IssueInstanceFixInfo
    ) -> None:
        assert (
            instance.id.local_id not in self._issue_instance_fix_info
        ), "Instance fix info already exists"
        self._issue_instance_fix_info[instance.id.local_id] = fix_info

    def get_text(self, shared_text_id: DBID) -> str:
        return self._shared_texts[shared_text_id.local_id].contents

    def get_shared_text_by_local_id(self, shared_text_id: int) -> SharedText:
        return self._shared_texts[shared_text_id]

    def get_shared_text(
        self, kind: SharedTextKind, content: str
    ) -> Optional[SharedText]:
        if kind in self._shared_text_lookup:
            contents = self._shared_text_lookup[kind]
            if content in contents and contents[content] in self._shared_texts:
                return self._shared_texts[contents[content]]
        return None

    def has_trace_frames_with_caller(
        self, kind: TraceKind, caller_id: DBID, caller_port: str
    ) -> bool:
        if self._trace_frames_map[kind]:
            key = (caller_id.local_id, caller_port)
            return key in self._trace_frames_map[kind]
        else:
            return False

    def has_postconditions_with_caller(self, caller_id: DBID, caller_port: str) -> bool:
        return self.has_trace_frames_with_caller(
            TraceKind.postcondition, caller_id, caller_port
        )

    def has_preconditions_with_caller(self, caller_id: DBID, caller_port: str) -> bool:
        return self.has_trace_frames_with_caller(
            TraceKind.precondition, caller_id, caller_port
        )

    def add_trace_annotation(self, annotation: TraceFrameAnnotation) -> None:
        self._trace_annotations[annotation.id.local_id] = annotation

    def get_condition_annotations(self, cond_id: int) -> List[TraceFrameAnnotation]:
        return [
            t
            for t in self._trace_annotations.values()
            if t.trace_frame_id.local_id == cond_id
        ]

    def get_annotation_trace_frames(self, ann_id: int) -> List[TraceFrame]:
        if ann_id in self._trace_frame_annotation_trace_frame_assoc:
            return [
                self.get_trace_frame_from_id(tf_id)
                for tf_id in self._trace_frame_annotation_trace_frame_assoc[ann_id]
            ]
        else:
            return []

    def add_trace_frame(self, trace_frame: TraceFrame) -> None:
        key = (trace_frame.caller_id.local_id, trace_frame.caller_port)
        rev_key = (trace_frame.callee_id.local_id, trace_frame.callee_port)
        # pyre-fixme[6]: Expected `TraceKind` for 1st param but got `str`.
        self._trace_frames_map[trace_frame.kind][key].add(trace_frame.id.local_id)
        # pyre-fixme[6]: Expected `TraceKind` for 1st param but got `str`.
        self._trace_frames_rev_map[trace_frame.kind][rev_key].add(
            trace_frame.id.local_id
        )
        self._trace_frames[trace_frame.id.local_id] = trace_frame

    def get_trace_frames_from_caller(
        self, kind: TraceKind, caller_id: DBID, caller_port: str
    ) -> List[TraceFrame]:
        key = (caller_id.local_id, caller_port)
        return [
            self._trace_frames[trace_frame_id]
            for trace_frame_id in self._trace_frames_map[kind][key]
        ]

    def get_trace_frame_from_id(self, id: int) -> TraceFrame:
        return self._trace_frames[id]

    def add_shared_text(self, shared_text: SharedText) -> None:
        assert (
            shared_text.id.local_id not in self._shared_texts
        ), "Shared text already exists"
        assert (
            shared_text.kind not in self._shared_text_lookup
            or shared_text.contents not in self._shared_text_lookup[shared_text.kind]
        ), "Shared text with same kind, contents exists"

        self._shared_texts[shared_text.id.local_id] = shared_text

        # Allow look up of SharedTexts by name and kind (to optimize
        # get_shared_text which is called when parsing each issue instance)
        self._shared_text_lookup[shared_text.kind][
            shared_text.contents
        ] = shared_text.id.local_id

    def get_or_add_shared_text(self, kind: SharedTextKind, name: str) -> SharedText:
        name = name[:SHARED_TEXT_LENGTH]
        shared_text = self.get_shared_text(kind, name)
        if shared_text is None:
            shared_text = SharedText.Record(id=DBID(), contents=name, kind=kind)
            self.add_shared_text(shared_text)
        return shared_text

    def add_trace_frame_leaf_assoc(
        self, trace_frame: TraceFrame, leaf: SharedText, depth: Optional[int]
    ) -> None:
        self._trace_frame_leaf_assoc[trace_frame.id.local_id][leaf.id.local_id] = depth

    def add_trace_frame_leaf_by_local_id_assoc(
        self, trace_frame: TraceFrame, leaf_id: int, depth: Optional[int]
    ) -> None:
        self._trace_frame_leaf_assoc[trace_frame.id.local_id][leaf_id] = depth

    def get_trace_frame_leaf_ids(self, trace_frame: TraceFrame) -> Set[int]:
        ids: Set[int] = {
            id for id in self._trace_frame_leaf_assoc[trace_frame.id.local_id]
        }
        return ids

    def get_trace_frame_leaf_ids_by_kind(
        self, trace_frame: TraceFrame, kind: SharedTextKind
    ) -> Set[int]:
        return {
            id
            for id in self._trace_frame_leaf_assoc[trace_frame.id.local_id]
            if self._shared_texts[id].kind == kind
        }

    def get_trace_frame_leaf_ids_with_depths(
        self, trace_frame: TraceFrame
    ) -> LeafIDToDepthMap:
        return self._trace_frame_leaf_assoc[trace_frame.id.local_id]

    def add_issue_instance_trace_frame_assoc(
        self, instance: IssueInstance, trace_frame: TraceFrame
    ) -> None:
        self._issue_instance_trace_frame_assoc[instance.id.local_id].add(
            trace_frame.id.local_id
        )
        self._trace_frame_issue_instance_assoc[trace_frame.id.local_id].add(
            instance.id.local_id
        )

    def add_trace_frame_annotation_trace_frame_assoc(
        self, annotation: TraceFrameAnnotation, trace_frame: TraceFrame
    ) -> None:
        self._trace_frame_annotation_trace_frame_assoc[annotation.id.local_id].add(
            trace_frame.id.local_id
        )
        self._trace_frame_trace_frame_annotation_assoc[trace_frame.id.local_id].add(
            annotation.id.local_id
        )

    def get_issue_instance_trace_frames(
        self, instance: IssueInstance
    ) -> List[TraceFrame]:
        if instance.id.local_id in self._issue_instance_trace_frame_assoc:
            return [
                self.get_trace_frame_from_id(id)
                for id in self._issue_instance_trace_frame_assoc[instance.id.local_id]
            ]
        else:
            return []

    def get_next_trace_frames(self, trace_frame: TraceFrame) -> Iterable[TraceFrame]:
        return self.get_trace_frames_from_caller(
            # pyre-fixme[6]: Expected `TraceKind` for 1st param but got `str`.
            trace_frame.kind,
            trace_frame.callee_id,
            trace_frame.callee_port,
        )

    def add_issue_instance_shared_text_assoc_id(
        self, instance: IssueInstance, shared_text_id: int
    ) -> None:
        self._issue_instance_shared_text_assoc[instance.id.local_id].add(shared_text_id)
        self._shared_text_issue_instance_assoc[shared_text_id].add(instance.id.local_id)

    def add_issue_instance_shared_text_assoc(
        self, instance: IssueInstance, shared_text: SharedText
    ) -> None:
        self.add_issue_instance_shared_text_assoc_id(instance, shared_text.id.local_id)

    def get_issue_instance_shared_texts(
        self, instance_id: int, kind: SharedTextKind
    ) -> List[SharedText]:
        return [
            self._shared_texts[msg_id]
            for msg_id in self._issue_instance_shared_text_assoc[instance_id]
            if self._shared_texts[msg_id].kind == kind
        ]

    def update_bulk_saver(self, bulk_saver: BulkSaver) -> None:
        bulk_saver.add_all(list(self._issues.values()))
        bulk_saver.add_all(list(self._issue_instances.values()))
        bulk_saver.add_all(list(self._trace_frames.values()))
        bulk_saver.add_all(list(self._issue_instance_fix_info.values()))
        bulk_saver.add_all(list(self._trace_annotations.values()))
        bulk_saver.add_all(list(self._shared_texts.values()))

        self._save_issue_instance_trace_frame_assoc(bulk_saver)
        self._save_trace_frame_leaf_assoc(bulk_saver)
        self._save_issue_instance_shared_text_assoc(bulk_saver)
        self._save_trace_frame_annotation_trace_frame_assoc(bulk_saver)

    def _save_issue_instance_trace_frame_assoc(self, bulk_saver: BulkSaver) -> None:
        for (
            trace_frame_id,
            instance_ids,
        ) in self._trace_frame_issue_instance_assoc.items():
            for instance_id in instance_ids:
                bulk_saver.add_issue_instance_trace_frame_assoc(
                    self._issue_instances[instance_id],
                    self._trace_frames[trace_frame_id],
                )

    def _save_trace_frame_annotation_trace_frame_assoc(
        self, bulk_saver: BulkSaver
    ) -> None:
        for (
            trace_annotation_id,
            trace_frame_ids,
        ) in self._trace_frame_annotation_trace_frame_assoc.items():
            for trace_frame_id in trace_frame_ids:
                bulk_saver.add_trace_frame_annotation_trace_frame_assoc(
                    self._trace_annotations[trace_annotation_id],
                    self._trace_frames[trace_frame_id],
                )

    def _save_trace_frame_leaf_assoc(self, bulk_saver: BulkSaver) -> None:
        """Adds trace frame leaf assocs to bulk saver after filtering them:
        1. if frame is a leaf, include all kinds
        2. otherwise, find outgoing leaf kinds and intersect with union of incoming
           leaf kinds of all successor frames.
        3. include only kinds that map to one of these outgoing kinds.
        """
        for trace_frame_id, leaf_ids in self._trace_frame_leaf_assoc.items():
            frame = self._trace_frames[trace_frame_id]
            valid_frame_leaf_ids = self._compute_valid_frame_leaves(frame)
            for leaf_id, depth in leaf_ids.items():
                leaf_text = self._shared_texts[leaf_id]
                if (
                    leaf_text.kind is SharedTextKind.FEATURE
                    or leaf_id in valid_frame_leaf_ids
                    or self._is_opposite_leaf(frame, leaf_text)
                ):
                    bulk_saver.add_trace_frame_leaf_assoc(leaf_text, frame, depth)
                else:
                    # Logging all the leaf kinds that are omitted causes large logs.
                    pass

    def _is_opposite_leaf(self, frame: TraceFrame, leaf: SharedText) -> bool:
        """We may be propagating sources along sink traces or vice versa. These should
        not be filtered and are identified here."""
        return (
            frame.kind == TraceKind.PRECONDITION and leaf.kind == SharedTextKind.SOURCE
        ) or (
            frame.kind == TraceKind.POSTCONDITION and leaf.kind == SharedTextKind.SINK
        )

    def _compute_valid_frame_leaves(self, frame: TraceFrame) -> Set[int]:
        leaf_mapping = frame.leaf_mapping
        is_leaf_frame = self.is_leaf_port(frame.callee_port)
        if not is_leaf_frame:
            callee_frames = self.get_next_trace_frames(frame)
            callee_leaf_ids = {
                callee_map.caller_leaf
                for callee_frame in callee_frames
                for callee_map in callee_frame.leaf_mapping
            }
        else:
            callee_leaf_ids = set()
        return {
            leaf_map.transform
            for leaf_map in leaf_mapping
            if is_leaf_frame or (leaf_map.callee_leaf in callee_leaf_ids)
        }

    def is_leaf_port(self, port: str) -> bool:
        return (
            port == "leaf"
            or port == "source"
            or port == "sink"
            or port.startswith("anchor:")
            or port.startswith("producer:")
        )

    def _save_issue_instance_shared_text_assoc(self, bulk_saver: BulkSaver) -> None:
        for (
            shared_text_id,
            instance_ids,
        ) in self._shared_text_issue_instance_assoc.items():
            for instance_id in instance_ids:
                bulk_saver.add_issue_instance_shared_text_assoc(
                    self._issue_instances[instance_id],
                    self._shared_texts[shared_text_id],
                )

    def compute_next_leaf_kinds(
        self, leaves: Set[int], leaf_mapping: Set[LeafMapping]
    ) -> Set[int]:
        """Normally, we would just intersect leaves and frame leaves, but since frame
        leaves can indicate local transforms of the form T1:...:Tn@G1..Gm:S, we need
        to be more careful.

        We first need to identify which frame leaves match by substituting the @
        for :, in general that would be T1:..Tn:G1..Gm:S. Then given these
        matches, erase everything up to and including the @ sign. That will be
        the new leaf kind.  In general, that is G1..Gm:S.

        leaf_mapping is already normalized to (caller_leaf_id, callee_leaf_id),
        i.e. which callers map to which caller ids obtained by performing the
        substitutions.

        For non-transform kinds, the leaf mapping contains identical
        caller_leaf_id, callee_leaf_id.
        """

        next_kinds = set()
        for leaf_map in leaf_mapping:
            if leaf_map.caller_leaf in leaves:
                next_kinds.add(leaf_map.callee_leaf)
        return next_kinds

    def compute_prev_leaf_kinds(
        self, leaves: Set[int], leaf_mapping: Set[LeafMapping]
    ) -> Set[int]:
        """Same as next_leaf_kinds but when following from leaves to issues."""
        next_kinds = set()
        for leaf_map in leaf_mapping:
            if leaf_map.callee_leaf in leaves:
                next_kinds.add(leaf_map.caller_leaf)
        return next_kinds

    def get_transform_normalized_leaf(self, leaf: str) -> str:
        return leaf.replace("@", ":", 1)

    def get_transform_normalized_kind_id(self, leaf_kind: SharedText) -> int:
        assert (
            leaf_kind.kind == SharedTextKind.SINK
            or leaf_kind.kind == SharedTextKind.SOURCE
        )
        if "@" in leaf_kind.contents:
            normal_name = self.get_transform_normalized_leaf(leaf_kind.contents)
            normal_kind = self.get_or_add_shared_text(leaf_kind.kind, normal_name)
            return normal_kind.id.local_id
        else:
            return leaf_kind.id.local_id

    def get_transformed_kind_id(self, leaf_kind: SharedText) -> int:
        assert (
            leaf_kind.kind == SharedTextKind.SINK
            or leaf_kind.kind == SharedTextKind.SOURCE
        )
        if "@" in leaf_kind.contents:
            splits = leaf_kind.contents.split("@", 1)
            remaining_kind = self.get_or_add_shared_text(leaf_kind.kind, splits[1])
            return remaining_kind.id.local_id
        else:
            return leaf_kind.id.local_id

    def get_callee_leaf_kinds_of_frame(self, trace_frame: TraceFrame) -> Set[int]:
        """Get leaf kinds expected by frames starting with the callee. In case of
        transforms, this is the untransformed kinds.
        """
        leaf_mapping = trace_frame.leaf_mapping
        assert leaf_mapping is not None
        return {leaf_map.callee_leaf for leaf_map in leaf_mapping}

    def get_caller_leaf_kinds_of_frame(self, trace_frame: TraceFrame) -> Set[int]:
        """Get leaf kinds expected by frames starting with the caller. In case of
        transforms, this is the transformed kinds.
        """
        leaf_mapping = trace_frame.leaf_mapping
        assert leaf_mapping is not None
        return {leaf_map.caller_leaf for leaf_map in leaf_mapping}



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from typing import Dict, Type

import click

from .cli_lib import commands, common_options
from .context import Context
from .db import DB, DBType
from .lint import lint
from .pipeline.base_parser import BaseParser
from .pipeline.mariana_trench_parser import Parser as MarianaTrenchParser
from .pipeline.pysa_taint_parser import Parser as PysaParser

LOG: logging.Logger = logging.getLogger("sapp")

PARSERS: Dict[str, Type[BaseParser]] = {
    "pysa": PysaParser,
    "mariana-trench": MarianaTrenchParser,
}


@common_options
@click.option(
    "--database-engine",
    "--database",
    type=click.Choice([DBType.SQLITE, DBType.MEMORY]),
    default=DBType.SQLITE,
    help="database engine to use",
)
@click.option(
    "--tool",
    type=click.Choice(list(PARSERS.keys())),
    default="pysa",
    help="tool the data is coming from",
)
@click.pass_context
def cli(
    ctx: click.Context,
    repository: str,
    database_name: str,
    database_engine: str,
    tool: str,
) -> None:
    ctx.obj = Context(
        repository=repository,
        database=DB(
            database_engine, os.path.expanduser(database_name), assertions=True
        ),
        tool=tool,
        parser_class=PARSERS[tool],
    )
    LOG.debug(f"Context: {ctx.obj}")


for command in commands:
    cli.add_command(command)
cli.add_command(lint)

if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s [%(levelname)s] %(message)s")
    cli()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, List, Tuple, Optional

import graphene
import sqlalchemy
from flask.views import View
from sqlalchemy.orm import Session

from .. import models
from ..context import Context
from ..db import DB
from ..filter import StoredFilter, FilterRecord
from ..models import DBID, Run, RunStatus
from ..sarif import SARIF
from .issues import Instance

if TYPE_CHECKING:
    from pathlib import Path  # usort: skip. Wants trailing whitespace

    from .issues import IssueQueryResult  # noqa

LOG: logging.Logger = logging.getLogger(__name__)


class Filter(graphene.ObjectType):
    name = graphene.String(required=True)
    description = graphene.String()
    json = graphene.String()

    @staticmethod
    def from_record(record: FilterRecord) -> Filter:
        return Filter(
            name=record.name, description=record.description, json=record.json
        )


def all_filters(session: Session) -> List[Filter]:
    return [Filter.from_record(record) for record in session.query(FilterRecord).all()]


def save_filter(session: Session, filter: Filter) -> None:

    existing = (
        session.query(FilterRecord).filter(FilterRecord.name == filter.name).first()
    )

    # pyre-ignore[6]: graphene too dynamic.
    filter_json = json.loads(filter.json)
    filter_json.pop("name", None)
    filter_json.pop("description", None)

    filter_record = StoredFilter(
        # pyre-ignore[6]: graphene too dynamic.
        filter.name,
        # pyre-ignore[6]: graphene too dynamic.
        filter.description or "",
        **filter_json,
    ).to_record()

    if not existing:
        session.add(filter_record)
        LOG.debug(f"Adding {filter}")
    else:
        existing.name = filter_record.name
        existing.description = filter_record.description
        existing.json = filter_record.json
        LOG.debug(f"Updating {filter}")

    session.commit()


class EmptyDeletionError(Exception):
    pass


def delete_filter(session: Session, name: str) -> None:
    deleted_rows = (
        session.query(FilterRecord).filter(FilterRecord.name == name).delete()
    )
    if deleted_rows == 0:
        raise EmptyDeletionError(f'No filter with `name` "{name}" exists.')
    LOG.info(f"Deleting {name}")
    session.commit()


def import_filter_from_path(database: DB, input_filter_path: Path) -> None:
    filter_paths = (
        list(input_filter_path.glob("**/*.json"))
        if input_filter_path.is_dir()
        else [input_filter_path]
    )

    imported_filterrecords = []
    for path in filter_paths:
        filter_instance = StoredFilter.from_file(path)
        imported_filterrecords.append(
            FilterRecord(
                name=filter_instance.name,
                description=filter_instance.description,
                json=filter_instance.to_json(),
            )
        )

    # TODO(T89343050)
    models.create(database)
    with database.make_session() as session:
        with session.begin_nested():
            for record in imported_filterrecords:
                session.merge(record)
                LOG.info(f"`{record.name}` filter has been imported")
        try:
            session.commit()
        except sqlalchemy.exc.DatabaseError:
            LOG.error(
                "Error: Database disk image is malformed. Please recreate your SAPP db"
            )
            raise


def delete_filters(database: DB, filter_names: Tuple[str]) -> None:
    if len(filter_names) <= 0:
        return

    with database.make_session() as session:
        for name in filter_names:
            if name == "":
                LOG.warning("You have provided an empty string for your filter name.")
                continue
            try:
                delete_filter(session, name)
            except EmptyDeletionError as error:
                LOG.exception(error)


class FilterNotFound(Exception):
    pass


def export_filter(
    database: DB, filter_name: str, output_filter_path: Optional[Path] = None
) -> None:
    with database.make_session() as session:
        try:
            record = (
                session.query(FilterRecord)
                .filter(FilterRecord.name == filter_name)
                .one_or_none()
            )
            if not record:
                raise FilterNotFound(
                    f"`{filter_name}` does not exist in `{database.dbname}`"
                )
            storedfilter_instance = StoredFilter.from_record(record)
            if output_filter_path:
                output_filter_path.write_text(storedfilter_instance.to_file())
                LOG.info(f"`{filter_name}` has been exported to `{output_filter_path}`")
            else:
                print(storedfilter_instance.to_file())
        except sqlalchemy.exc.OperationalError:
            LOG.error(
                "Error: Database disk image is malformed or "
                "database has not been initialized properly."
                "Please recreate your SAPP db."
            )
            raise


class ServeExportFilter(View):
    def __init__(self, session: Session) -> None:
        self.session = session

    # pyre-fixme[14]: Inconsistent override
    def dispatch_request(self, filter_name: str) -> str:
        try:
            record = (
                self.session.query(FilterRecord)
                .filter(FilterRecord.name == filter_name)
                .one_or_none()
            )
            if not record:
                raise FilterNotFound(f"`{filter_name}` does not exist")
            storedfilter_instance = StoredFilter.from_record(record)
            return storedfilter_instance.to_file()
        except sqlalchemy.exc.OperationalError:
            LOG.error(
                "Error: Database disk image is malformed or "
                "database has not been initialized properly."
                "Please recreate your SAPP db."
            )
            raise


class InvalidRunException(Exception):
    pass


def filter_run(
    context: Context,
    run_id_input: int,
    filter_path: Path,
    output_format: str,
) -> None:
    with context.database.make_session() as session:
        run_id: Run = (
            session.query(Run)
            .filter(Run.status == RunStatus.FINISHED)
            .filter(Run.id == run_id_input)
            .scalar()
        )
        if run_id is None:
            raise InvalidRunException(
                f"No finished run with ID {run_id_input} exists. Make sure you have run 'sapp analyze' before running `sapp filter`"
            )

        paths = (
            list(filter_path.glob("**/*.json"))
            if filter_path.is_dir()
            else [filter_path]
        )
        filter_instances = [StoredFilter.from_file(path) for path in paths]

        if len(filter_instances) <= 0:
            LOG.error(f"No valid filters found in `{filter_path}`")
            return

        query_results = set()
        for filter_instance in filter_instances:
            query_result = (
                Instance(session, DBID(run_id_input))
                .where_filter(filter_instance)
                .get()
            )
            LOG.info(
                (
                    f"Applying `{filter_instance.name}` to run `{run_id_input}` "
                    f"resulted in {len(query_result)} issues"
                )
            )
            for issue in query_result:
                query_results.add(issue)

        total_filtered_issues_output = (
            f"Total number of issues after filtering: {len(query_results)}"
        )
        if len(query_results) <= 0:
            LOG.error(total_filtered_issues_output)
            return
        else:
            LOG.info(total_filtered_issues_output)
        if output_format == "sapp":
            output_json = {"issues": [issue.to_json() for issue in query_results]}
            print(json.dumps(output_json, indent=2))
        elif output_format == "sarif":
            sarif_output = SARIF(context.tool, session, query_results)
            print(sarif_output.to_json())



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from multiprocessing import Pool
from typing import Iterable, List, Tuple, Type, Union

from ..analysis_output import AnalysisOutput, Metadata
from . import ParseCondition, ParseConditionTuple, ParseIssue, ParseIssueTuple
from .base_parser import BaseParser

log: logging.Logger = logging.getLogger("sapp")
logging.basicConfig(format="%(asctime)s [%(levelname)s] %(message)s")


# We are going to call this per process, so we need to pass in and return
# serializable data. And as a single arg, as far as I can tell. Which is why the
# args type looks so silly.
def parse(
    args: Tuple[Tuple[Type[BaseParser], List[str], Metadata], str]
) -> List[Union[ParseCondition, ParseIssue, ParseConditionTuple, ParseIssueTuple]]:
    (base_parser, repo_dirs, metadata), path = args

    parser = base_parser(repo_dirs)
    parser.initialize(metadata)

    with open(path) as handle:
        return list(parser.parse_handle(handle))


class ParallelParser(BaseParser):
    def __init__(self, parser_class: Type[BaseParser], repo_dirs: List[str]) -> None:
        super().__init__(repo_dirs)
        self.parser: Type[BaseParser] = parser_class

    def parse(
        self, input: AnalysisOutput
    ) -> Iterable[
        Union[ParseCondition, ParseIssue, ParseConditionTuple, ParseIssueTuple]
    ]:
        log.info("Parsing in parallel")
        files = list(input.file_names())

        # Pair up the arguments with each file.
        num_files = len(files)
        args = zip([(self.parser, self.repo_dirs, input.metadata)] * num_files, files)

        with Pool(processes=None) as pool:
            for idx, f in enumerate(pool.imap_unordered(parse, args)):
                if idx % 10 == 0:
                    cur = idx + 1
                    pct = round((cur / num_files) * 100, 2)
                    log.info(f"{cur}/{num_files} ({pct}%) files parsed")
                yield from f



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import fnmatch
import os
import re


class ShardedFileComponents(object):
    """
    Splits a sharded file name or pattern into its components:
    - directory
    - stem
    - extension
    - shard_index
    - shard_total
    For shard patterns @n, the shard index is set to -1.
    For shard patterns @*, the shard total is also set to -1.
    """

    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, filepattern) -> None:
        # pyre-fixme[4]: Attribute must be annotated.
        self.directory, root = os.path.split(filepattern)
        m = re.match(r"([^@]+)@([^.@]+)(\.[^.@]*)?$", root)
        if not m:
            raise ValueError("Not a sharded file: {}".format(filepattern))

        # pyre-fixme[4]: Attribute must be annotated.
        # pyre-fixme[58]: `>=` is not supported for operand types `Optional[int]`
        #  and `int`.
        self.extension = m.group(3) if m.lastindex >= 3 else ""
        # pyre-fixme[4]: Attribute must be annotated.
        # pyre-fixme[58]: `>=` is not supported for operand types `Optional[int]`
        #  and `int`.
        self.stem = m.group(1) if m.lastindex >= 1 else ""

        # pyre-fixme[58]: `>=` is not supported for operand types `Optional[int]`
        #  and `int`.
        shards = m.group(2) if m.lastindex >= 2 else ""
        if shards == "*":
            # pyre-fixme[4]: Attribute must be annotated.
            self.shard_index = -1
            # pyre-fixme[4]: Attribute must be annotated.
            self.shard_total = -1
        else:
            try:
                self.shard_total = int(shards)
                self.shard_index = -1
            except ValueError:
                # Look for ?????-of-????? pattern.
                m = re.match(r"(\d{5})-of-(\d{5})$", shards)
                if not m:
                    raise ValueError("Invalid shard specification: {}".format(shards))
                self.shard_index = int(m.group(1))
                self.shard_total = int(m.group(2))

        if self.directory == "":
            self.directory = "."

        # Sanity check shard total.
        if self.shard_total == 0:
            raise ValueError("Invalid shard total 0")

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def get_shard_filename(self, index):
        if index < 0 or index >= self.shard_total:
            raise ValueError("Invalid shard index")
        return os.path.join(
            self.directory,
            "{}@{:05d}-of-{:05d}{}".format(
                self.stem, index, self.shard_total, self.extension
            ),
        )

    # pyre-fixme[3]: Return type must be annotated.
    def is_at_n_pattern(self):
        return self.shard_total > 0 and self.shard_index == -1

    # pyre-fixme[3]: Return type must be annotated.
    def is_at_star_pattern(self):
        return self.shard_total == -1 and self.shard_index == -1


class ShardedFile(object):
    """A sharded file object represents a set of files sharded according to the
    following pattern:

         path/stem@?????-of-?????.ext

    The ????? represent 0-padded decimal numbers. For example:
    1) path/foo@00000-of-00001.ext is the single shard of a 1-sharded file.
    1) path/foo@00000-of-00002.ext and path/foo@00001-of-00002.ext are the
       two shards of a 2-sharded file.

    There are 2 ways to name a sharded file set:
    1) path/foo@*.ext -- match a sharded file with arbitrary # of shards.
    2) path/foo@n.ext -- match a sharded file with n shards.

    A sharded file object is constructed by giving it a file pattern of the
    above form. The pattern is checked and the appropriate shards found. Errors
    are indicated as properties on the sharded file.

    The list of shards is available in shard_file_names.
    """

    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, pattern) -> None:
        """
        Determine shards from pattern and record errors.
        """
        comps = ShardedFileComponents(pattern)
        if comps.is_at_star_pattern():
            comps.shard_total = self._find_unambiguous_shard_total(comps, pattern)
        elif not comps.is_at_n_pattern():
            raise ValueError("Pattern should use @n or @* for shard specification.")

        # now we have an @n spec.
        self._set_shard_file_names(comps)
        self._shard_file_names.sort()

    # pyre-fixme[3]: Return type must be annotated.
    def get_filenames(self):
        return self._shard_file_names

    # pyre-fixme[2]: Parameter must be annotated.
    def _set_shard_file_names(self, pcomps) -> None:
        self._shard_file_names = []
        for i in range(pcomps.shard_total):
            filename = pcomps.get_shard_filename(i)
            if not os.path.isfile(filename):
                raise ValueError("Shard {} does not exist.".format(filename))
            self._shard_file_names.append(filename)

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _find_unambiguous_shard_total(self, pcomps, pattern):
        dir = pcomps.directory
        if not os.path.isdir(dir):
            raise ValueError("Not a directory {}".format(dir))

        pattern = pcomps.stem + "@?????-of-?????" + pcomps.extension
        seen_shard_count = -1
        for file in os.listdir(dir):
            if fnmatch.fnmatch(file, pattern):
                try:
                    comps = ShardedFileComponents(file)
                except ValueError:
                    continue
                if seen_shard_count == -1:
                    seen_shard_count = comps.shard_total
                elif seen_shard_count != comps.shard_total:
                    [a, b] = sorted([seen_shard_count, comps.shard_total])
                    raise ValueError(
                        "@* matches ambiguous shard sets: @{} and @{}".format(a, b)
                    )
        if seen_shard_count == -1:
            raise ValueError("Pattern matches no sharded file set: {}".format(pattern))
        return seen_shard_count



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import logging
from collections import namedtuple
from itertools import tee
from typing import Dict, List, Optional, Set, Tuple, Type

from munch import Munch
from sqlalchemy import Column, String, and_, exc, inspect, or_, types
from sqlalchemy.dialects import mysql
from sqlalchemy.dialects.mysql import BIGINT
from sqlalchemy.orm import Session

from .iterutil import split_every

log: logging.Logger = logging.getLogger("sapp")


"""Number of variables that can safely be set on a single DB call"""
BATCH_SIZE = 450


# The following three DBID classes require some explanation. Normally models
# will reference each other by their id. But we do bulk insertion at the end
# of our processing, which means the id isn't set until later. Having a DBID
# object allows these models to reference each other before that point. When
# we are ready to insert into the database, PrimaryKeyGenerator will give it
# an ID. Any other models referencing that DBID object will now be able to use
# the real id.


class DBID(object):
    __slots__ = ["_id", "is_new", "local_id"]

    # Temporary IDs that are local per run (local_id) are assigned for each
    # DBID object on creation. This acts as a key for the object in map-like
    # structures of DB objects without having to define a hashing function for
    # each of them. next_id tracks the next available int to act as an id.
    next_id: int = 0

    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, id=None) -> None:
        self.resolve(id)
        self.local_id: int = DBID.next_id
        DBID.next_id += 1

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def resolve(self, id, is_new=True):
        self._check_type(id)
        self._id = id
        self.is_new = is_new
        return self

    # pyre-fixme[3]: Return type must be annotated.
    def resolved(self):
        id = self._id

        # We allow one level of a DBID pointing to another DBID
        if isinstance(id, DBID):
            id = id.resolved()

        return id

    # pyre-fixme[2]: Parameter must be annotated.
    def _check_type(self, id) -> None:
        if not isinstance(id, (int, type(None), DBID)):
            raise TypeError(
                "id expected to be type '{}' but was type '{}'".format(int, type(id))
            )

    # Allow DBIDs to be added and compared as ints
    # pyre-fixme[3]: Return type must be annotated.
    def __int__(self):
        return self.resolved()

    # pyre-fixme[3]: Return type must be annotated.
    def __str__(self):
        return str(self.resolved())

    # pyre-fixme[2]: Parameter must be annotated.
    def __add__(self, other) -> int:
        return int(self) + int(other)

    # pyre-fixme[2]: Parameter must be annotated.
    def __lt__(self, other) -> bool:
        return int(self) < int(other)

    # pyre-fixme[2]: Parameter must be annotated.
    def __gt__(self, other) -> bool:
        return int(self) > int(other)

    # pyre-fixme[2]: Parameter must be annotated.
    def __ge__(self, other) -> bool:
        return int(self) >= int(other)

    # pyre-fixme[2]: Parameter must be annotated.
    def __le__(self, other) -> bool:
        return int(self) <= int(other)

    def __repr__(self) -> str:
        return "<{}(id={}) object at 0x{:x}>".format(
            self.__class__.__name__, self._id, id(self)
        )


class DBIDType(types.TypeDecorator):
    impl = types.Integer
    cache_ok = False

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def process_bind_param(self, value, dialect):
        # If it is a DBID wrapper, then write the contained value. Otherwise it
        # may be resolved already, or None.
        if isinstance(value, DBID):
            return value.resolved()
        else:
            return value

    # pyre-fixme[2]: parameter must be annotated.
    # pyre-fixme[2]: parameter must be annotated.
    def process_result_value(self, value, dialect) -> DBID:
        return DBID(value)

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def load_dialect_impl(self, dialect):
        if dialect.name == "mysql":
            return dialect.type_descriptor(mysql.INTEGER(unsigned=True))
        return self.impl


class BIGDBIDType(DBIDType):
    impl = types.BigInteger

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def load_dialect_impl(self, dialect):
        if dialect.name == "mysql":
            return dialect.type_descriptor(mysql.BIGINT(unsigned=True))
        return self.impl


class PrepareMixin(object):
    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def prepare(cls, session, pkgen, items):
        """This is called immediately before the items are written to the
        database. pkgen is passed in to allow last-minute resolving of ids.
        """
        for item in cls.merge(session, items):
            if hasattr(item, "id"):
                item.id.resolve(id=pkgen.get(cls), is_new=True)
            # pyre-fixme[16]: `PrepareMixin` has no attribute `to_dict`.
            yield cls.to_dict(item)

    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def merge(cls, session, items):
        """Models should override this to perform a merge"""
        return items

    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _merge_by_key(cls, session, items, attr):
        return cls._merge_by_keys(
            session, items, lambda item: getattr(item, attr.key), attr
        )

    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _merge_by_keys(cls, session, items, hash_item, *attrs):
        """An object can have multiple attributes as its key. This merges the
        items to be added with existing items in the database based on their
        key(s).

        session: Session object for querying the DB.
        items: Iterator of items to be added to the DB.
        hash_item: Function that takes as in put the item to be added and
                   returns a hash of it.
        attrs: List of attributes of the object/class that represent the
               object's key.

        Returns the next item (in items) that is not already in the DB.
        """
        # Note: items is an iterator, not an iterable, 'tee' is a must.
        items_iter1, items_iter2 = tee(items)

        keys = {}  # map of hash -> keys of the item
        for i in items_iter1:
            # An item's key is a map of 'attr -> item[attr]' where attr is
            # usually a column name.
            # For 'SharedText', its key would look like: {
            #   "kind": "feature",
            #   "contents": "via tito",
            # }
            item_hash = hash_item(i)
            keys[item_hash] = {attr.key: getattr(i, attr.key) for attr in attrs}

        # Find existing items.
        existing_ids = {}  # map of item_hash -> existing ID
        cls_attrs = [getattr(cls, attr.key) for attr in attrs]
        for fetch_keys in split_every(BATCH_SIZE, keys.values()):
            filters = []
            for fetch_key in fetch_keys:
                # Sub-filters for checking if item with fetch_key is in the DB
                # Example: [
                #   SharedText.kind.__eq__("feature"),
                #   SharedText.contents.__eq__("via tito"),
                # ]
                subfilter = [
                    getattr(cls, attr).__eq__(val) for attr, val in fetch_key.items()
                ]
                filters.append(and_(*subfilter))
            existing_items = (
                # pyre-fixme[16]: `PrepareMixin` has no attribute `id`.
                session.query(cls.id, *cls_attrs)
                .filter(or_(*(filters)))
                .all()
            )
            for existing_item in existing_items:
                item_hash = hash_item(existing_item)
                existing_ids[item_hash] = existing_item.id

        # Now see if we can merge
        new_items = {}
        for i in items_iter2:
            item_hash = hash_item(i)
            if item_hash in existing_ids:
                # The key is already in the DB
                i.id.resolve(existing_ids[item_hash], is_new=False)
            elif item_hash in new_items:
                # The key is already in the list of new items
                i.id.resolve(new_items[item_hash].id, is_new=False)
            else:
                # The key is new
                new_items[item_hash] = i
                yield i

    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _merge_assocs(cls, session, items, id1, id2):
        new_items = {}
        for i in items:
            r1 = getattr(i, id1.key)
            r2 = getattr(i, id2.key)
            key = (r1.resolved(), r2.resolved())
            if key not in new_items:
                new_items[key] = i
                yield i


# The record mixin class is more efficient than the MutableRecordMixin, so it
# should be preferred. But the performance isn't from the mutability, it's
# because we use namedtuples, which creates a new class on demand, which uses
# __slots__, which is more efficient. Both of these mixins can be replaced when
# we have dynamically created classes with the slots set. But until then,
# prefer RecordMixin unless you need to change fields after creation.
class RecordMixin(object):
    # pyre-fixme[4]: Attribute must be annotated.
    _record = None

    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def Record(cls, extra_fields=None, **kwargs):
        if not cls._record:
            if not extra_fields:
                extra_fields = []
            mapper = inspect(cls)
            keys = [c.key for c in mapper.column_attrs] + ["model"] + extra_fields
            cls._record = namedtuple(cls.__name__ + "Record", keys)

        return cls._record(model=cls, **kwargs)

    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def to_dict(cls, obj):
        return obj._asdict()


class MutableRecordMixin(object):
    @classmethod
    # pyre-fixme[2]: Parameter must be annotated.
    def Record(cls, **kwargs) -> Munch:
        return Munch(model=cls, **kwargs)

    @classmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def to_dict(cls, obj):
        return obj.toDict()


class PrimaryKeyBase(PrepareMixin, RecordMixin):  # noqa
    """Subclass this and include your declarative_base mixin"""

    __tablename__ = "primary_keys"

    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
    table_name: str = Column(
        String(length=100),
        doc="Name of the table that this row stores the next available primary key for",
        nullable=False,
        primary_key=True,
    )

    # pyre-fixme[8]: Attribute has type `int`; used as
    #  `Column[Variable[sqlalchemy.sql.type_api._U]]`.
    current_id: int = Column(
        BIGINT(unsigned=True).with_variant(BIGINT, "sqlite"),
        doc="The current/latest id used in the table.",
        nullable=False,
        primary_key=False,
    )


class PrimaryKeyGeneratorBase:  # pyre-ignore[13]
    """Keep track of DB objects' primary keys by ourselves rather than relying
    on SQLAlchemy, so we can supply them as arguments when creating association
    objects. Subclass to define PRIMARY_KEY class and QUERY_CLASSES."""

    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    PRIMARY_KEY: Type
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    QUERY_CLASSES: Set[Type]

    # Map from class name to an ID range (next_id, max_reserved_id)
    pks: Dict[str, Tuple[int, int]] = {}

    def reserve(
        self,
        session: Session,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        saving_classes: List[Type],
        item_counts: Optional[Dict[str, int]] = None,
    ) -> "PrimaryKeyGeneratorBase":
        """
        session - Session for DB operations.
        saving_classes - class objects that need to be saved e.g. Issue, Run
        item_counts - map from class name to the number of items, for preallocating
        id ranges
        """
        query_classes = {cls for cls in saving_classes if cls in self.QUERY_CLASSES}
        for cls in query_classes:
            if item_counts and cls.__name__ in item_counts:
                count = item_counts[cls.__name__]
            else:
                count = 1
            self._reserve_id_range(session, cls, count)

        return self

    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    def _lock_pk_with_retries(self, session: Session, cls: Type) -> Optional[Type]:
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        cls_pk: Optional[Type] = None
        retries: int = 6
        while retries > 0:
            try:
                cls_pk = (
                    session.query(self.PRIMARY_KEY)
                    .filter(self.PRIMARY_KEY.table_name == cls.__name__)
                    .with_for_update()
                    .first()
                )
                # if we're here, the record has been locked, or there is no record
                retries = 0
            except exc.OperationalError as ex:
                # Failed to get exclusive lock on the record, so we retry
                retries -= 1
                # Re-raise the exception if our retries are exhausted
                if retries == 0:
                    raise ex
        return cls_pk

    def _reserve_id_range(
        self,
        session: Session,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        cls: Type,
        count: int,
    ) -> None:
        cls_pk = self._lock_pk_with_retries(session, cls)
        if not cls_pk:
            # If cls_pk is None, then we query the data table for the max ID
            # and use that as the current_id in the primary_keys table. This
            # should only occur once (the except with a rollback means any
            # additional attempt will fail to add a row, and use the "current"
            # id value)
            row = session.query(cls.id).order_by(cls.id.desc()).first()
            try:
                session.execute(
                    "INSERT INTO primary_keys(table_name, current_id) \
                    VALUES (:table_name, :current_id)",
                    {
                        "table_name": cls.__name__,
                        "current_id": (row.id) if row else 0,
                    },
                )
                session.commit()
            except exc.SQLAlchemyError:
                session.rollback()
            cls_pk = self._lock_pk_with_retries(session, cls)

        if cls_pk:
            next_id = cls_pk.current_id + 1
            cls_pk.current_id = cls_pk.current_id + count
            pk_entry: Tuple[int, int] = (next_id, cls_pk.current_id)
            session.commit()
            self.pks[cls.__name__] = pk_entry

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def get(self, cls):
        assert cls in self.QUERY_CLASSES, (
            "%s primary key should be generated by SQLAlchemy" % cls.__name__
        )
        assert cls.__name__ in self.pks, (
            "%s primary key needs to be initialized before use" % cls.__name__
        )
        (pk, max_pk) = self.pks[cls.__name__]
        assert pk <= max_pk, "%s reserved primary key range exhausted" % cls.__name__
        self.pks[cls.__name__] = (pk + 1, max_pk)
        return pk



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

TYPECHECK_ERROR = 102


class AIException(Exception):
    pass


class AIRecoverableException(AIException):
    pass


class AIProcessException(AIRecoverableException):
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, message, error_code):
        super().__init__(message)
        # pyre-fixme[4]: Attribute must be annotated.
        self.error_code = error_code


class ParseTypeException(Exception):
    pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest import TestCase

from ..iterutil import split_every


class UtilsTest(TestCase):
    def test_split_every(self) -> None:
        self.assertEqual(
            list(split_every(2, range(10))), [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file defines the underlying db used by SAPP library.
"""

import logging
from contextlib import contextmanager
from typing import Any, Iterator, Optional, Type

import sqlalchemy
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session, scoped_session, sessionmaker
from sqlalchemy.pool import AssertionPool, Pool

from . import errors
from .decorators import retryable

LOG: logging.Logger = logging.getLogger("sapp")


class DBType(sqlalchemy.Enum):
    XDB = "xdb"  # not yet implemented
    SQLITE = "sqlite"
    MEMORY = "memory"


class DB(object):
    """Interact with the database type requested"""

    """File-based DB when using SQLITE"""
    DEFAULT_DB_FILE = "sapp.db"

    def __init__(
        self,
        dbtype: str,
        dbname: Optional[str] = None,
        debug: bool = False,
        read_only: bool = False,
        assertions: bool = False,
    ) -> None:
        self.dbtype = dbtype
        self.dbname: str = dbname or self.DEFAULT_DB_FILE
        self.debug = debug
        self.read_only = read_only
        self.assertions = assertions

        self.poolclass: Optional[Type[Pool]] = AssertionPool if assertions else None

        if dbtype == DBType.MEMORY:
            # pyre-fixme[4]: Attribute must be annotated.
            self.engine = sqlalchemy.create_engine(
                sqlalchemy.engine.url.URL("sqlite", database=":memory:"),
                echo=debug,
                poolclass=self.poolclass,
            )
        elif dbtype == DBType.SQLITE:
            self.engine = sqlalchemy.create_engine(
                sqlalchemy.engine.url.URL("sqlite", database=self.dbname),
                echo=debug,
                poolclass=self.poolclass,
            )
        elif dbtype == DBType.XDB:
            self._create_xdb_engine()
        else:
            raise errors.AIException(f"Invalid db type: {dbtype}")

    def _create_xdb_engine(self) -> None:
        raise NotImplementedError

    @contextmanager
    def make_session(self, *args: Any, **kwargs: Any) -> Iterator[Session]:
        session = self.make_session_object(*args, **kwargs)
        try:
            yield session
        finally:
            self.close_session(session)

    @retryable(num_tries=2, retryable_exs=[OperationalError])
    def make_session_object(self, *args: Any, **kwargs: Any) -> Session:
        # use scoped_session so sessionmaker generates the same session in
        # different threads. This is useful for UTs.
        session = scoped_session(sessionmaker(bind=self.engine))(*args, **kwargs)
        ping_db(session)
        if self.dbtype == DBType.XDB:
            # Make sure SQL doesn't quit on us after 10s. Sometimes merging data takes
            # longer.
            session.execute("SET SESSION wait_timeout = %d" % 30)

        return session

    @retryable(num_tries=2, retryable_exs=[OperationalError])
    def close_session(self, session: Session) -> None:
        session.close()


def ping_db(session: Session) -> None:
    session.execute("SELECT 1")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import datetime
from typing import Optional

from ..bulk_saver import BulkSaver
from ..models import (
    DBID,
    FrameReachability,
    Issue,
    IssueDBID,
    IssueInstance,
    IssueInstanceFixInfo,
    Run,
    SharedText,
    SharedTextKind,
    SourceLocation,
    TraceFrame,
    TraceKind,
)
from ..trace_graph import LeafMapping, TraceGraph


class FakeObjectGenerator:
    def __init__(self, graph: Optional[TraceGraph] = None, run_id=0):
        self.reinit(run_id)
        self.graph = graph

    def reinit(self, run_id):
        self.saver = BulkSaver()
        self.handle = 0
        self.source_name_id = 0
        self.sink_name_id = 0
        self.shared_text_name_id = 0
        self.run_id = run_id

    def save_all(self, db):
        if self.graph:
            self.graph.update_bulk_saver(self.saver)
        self.saver.save_all(db)
        self.saver = BulkSaver()

    def issue(
        self,
        filename="/r/lib/server/posts/request.py",
        callable="lib.server.posts.request.new_post_request",
        handle=None,
        code=None,
        status="uncategorized",
    ):
        self.handle += 1
        now = datetime.datetime.now()
        result = Issue.Record(
            id=IssueDBID(),
            handle=str(self.handle) if not handle else handle,
            callable_line=1,
            message="Tainted information may eventually flow...",
            code=code or (6015 + self.handle),
            filename=filename,
            callable=callable,
            first_seen=now,
            last_seen=now,
            line=1,
            start=1,
            end=2,
            status=status,
        )
        if self.graph:
            self.graph.add_issue(result)
        else:
            self.saver.add(result)
        return result

    def precondition(
        self,
        caller="double_meh",
        caller_port="at the end of universe",
        callee="triple_meh",
        callee_port="at the beginning of time",
        filename="lib/server/posts/request.py",
        location=(4, 5, 6),
        leaves=None,
        reachability=FrameReachability.UNREACHABLE,
    ):
        leaves = leaves or []
        filename_record = self.filename(filename)
        caller_record = self.callable(caller)
        callee_record = self.callable(callee)
        trace_frame = TraceFrame.Record(
            extra_fields=["leaf_mapping"],
            leaf_mapping={
                LeafMapping(leaf.id.local_id, leaf.id.local_id, leaf.id.local_id)
                for (leaf, _) in leaves
                if leaf.kind == SharedTextKind.source
                or leaf.kind == SharedTextKind.sink
            },
            id=DBID(),
            kind=TraceKind.PRECONDITION,
            caller_id=caller_record.id,
            caller_port=caller_port,
            callee_id=callee_record.id,
            callee_port=callee_port,
            callee_location=SourceLocation(location[0], location[1], location[2]),
            filename_id=filename_record.id,
            titos=[],
            run_id=self.run_id,
            type_interval_lower=5,
            type_interval_upper=7,
            preserves_type_context=False,
            reachability=reachability,
        )
        if self.graph:
            self.graph.add_trace_frame(trace_frame)
            for (leaf, depth) in leaves:
                self.graph.add_trace_frame_leaf_assoc(trace_frame, leaf, depth)
        else:
            self.saver.add(trace_frame)
        return trace_frame

    def postcondition(
        self,
        caller="quadruple_meh",
        caller_port="caller_meh",
        callee="quintuple_meh",
        callee_port="callee_meh",
        filename="lib/server/posts/response.py",
        location=(4, 5, 6),
        leaves=None,
    ):
        leaves = leaves or []
        filename_record = self.filename(filename)
        caller_record = self.callable(caller)
        callee_record = self.callable(callee)
        trace_frame = TraceFrame.Record(
            extra_fields=["leaf_mapping"],
            leaf_mapping={
                LeafMapping(leaf.id.local_id, leaf.id.local_id, leaf.id.local_id)
                for (leaf, _) in leaves
                if leaf.kind == SharedTextKind.source
                or leaf.kind == SharedTextKind.sink
            },
            id=DBID(),
            kind=TraceKind.POSTCONDITION,
            caller_id=caller_record.id,
            caller_port=caller_port,
            callee_id=callee_record.id,
            callee_port=callee_port,
            callee_location=SourceLocation(location[0], location[1], location[2]),
            filename_id=filename_record.id,
            titos=[],
            run_id=self.run_id,
            type_interval_lower=5,
            type_interval_upper=7,
            preserves_type_context=False,
            reachability=FrameReachability.UNREACHABLE,
        )
        if self.graph:
            self.graph.add_trace_frame(trace_frame)
            for (leaf, depth) in leaves:
                self.graph.add_trace_frame_leaf_assoc(trace_frame, leaf, depth)
        else:
            self.saver.add(trace_frame)
        return trace_frame

    def shared_text(self, contents, kind):
        if self.graph:
            shared_text = self.graph.get_shared_text(kind, contents)
            if shared_text is not None:
                return shared_text

        result = SharedText.Record(id=DBID(), contents=contents, kind=kind)
        if self.graph:
            self.graph.add_shared_text(result)
        else:
            self.saver.add(result)
        return result

    def run(self, differential_id=None, job_id=None, kind=None):
        self.run_id += 1
        # Not added to bulksaver or graph
        return Run(
            id=DBID(self.run_id),
            date=datetime.datetime.now(),
            hh_version="1234567890",
            revision_id=12345,
            differential_id=differential_id,
            job_id=job_id,
            kind=kind,
        )

    def feature(self, name="via:feature"):
        return self.shared_text(contents=name, kind=SharedTextKind.FEATURE)

    def source(self, name="source"):
        return self.shared_text(contents=name, kind=SharedTextKind.SOURCE)

    def source_detail(self, name="source_detail"):
        return self.shared_text(contents=name, kind=SharedTextKind.SOURCE_DETAIL)

    def sink(self, name="sink"):
        return self.shared_text(contents=name, kind=SharedTextKind.SINK)

    def sink_detail(self, name="sink_detail"):
        return self.shared_text(contents=name, kind=SharedTextKind.SINK_DETAIL)

    def filename(self, name="/r/some/filename.py"):
        return self.shared_text(contents=name, kind=SharedTextKind.FILENAME)

    def callable(self, name="Foo.barMethod"):
        return self.shared_text(contents=name, kind=SharedTextKind.CALLABLE)

    def message(self, name="this is bad"):
        return self.shared_text(contents=name, kind=SharedTextKind.MESSAGE)

    def instance(
        self,
        message="this is bad",
        filename="/r/some/filename.py",
        callable="Foo.barMethod",
        issue_id=None,
        min_trace_length_to_sources=None,
        min_trace_length_to_sinks=None,
    ):
        issue_id = issue_id if issue_id is not None else DBID(1)
        filename = self.filename(filename)
        message = self.message(message)
        callable = self.callable(callable)
        result = IssueInstance.Record(
            id=DBID(),
            location=SourceLocation(6, 7, 8),
            filename_id=filename.id,
            message_id=message.id,
            callable_id=callable.id,
            run_id=self.run_id,
            issue_id=issue_id,
            min_trace_length_to_sources=min_trace_length_to_sources,
            min_trace_length_to_sinks=min_trace_length_to_sinks,
        )
        if self.graph:
            self.graph.add_issue_instance(result)
        else:
            self.saver.add(result)
        return result

    def fix_info(self):
        result = IssueInstanceFixInfo.Record(id=DBID(), fix_info="fixthis")
        if self.graph:
            self.graph.add_fix_info(result)
        else:
            self.saver.add(result)
        return result



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from setuptools import find_packages, setup


def _requirements():
    return Path("requirements.txt").read_text()


setup(
    name="fb-sapp",
    version="0.5.1",
    description="Static Analysis Post-Processor for processing taint analysis results.",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=_requirements(),
    entry_points={"console_scripts": ["sapp = sapp.cli:cli"]},
    packages=find_packages(),
    url="https://github.com/facebook/sapp",
    author="Facebook",
    maintainer_email="pyre@fb.com",
    package_data={
        "sapp.ui": [
            "frontend/build/*",
            "frontend/build/static/css/*",
            "frontend/build/static/js/*",
        ],
    },
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Set, Tuple, Union

import graphene
from graphql.execution.base import ResolveInfo
from sqlalchemy.orm import Session, aliased
from sqlalchemy.orm.util import AliasedClass

from ..models import (
    DBID,
    IssueInstanceTraceFrameAssoc,
    SharedText,
    SharedTextKind,
    SourceLocation,
    TraceFrame,
    TraceFrameLeafAssoc,
    TraceKind,
)
from . import run

FilenameText: AliasedClass = aliased(SharedText)
CallableText: AliasedClass = aliased(SharedText)
CallerText: AliasedClass = aliased(SharedText)
CalleeText: AliasedClass = aliased(SharedText)
MessageText: AliasedClass = aliased(SharedText)


LEAF_NAMES: Set[str] = {"source", "sink", "leaf"}


class TraceFrameQueryResultType(graphene.ObjectType):
    class Meta:
        interfaces = (graphene.relay.Node,)

    frame_id = graphene.ID()
    caller = graphene.String()
    caller_port = graphene.String()
    callee = graphene.String()
    callee_port = graphene.String()
    caller_id = graphene.ID()
    callee_id = graphene.ID()
    callee_location = graphene.ID()
    kind = graphene.String()
    filename = graphene.String()
    trace_length = graphene.Int()
    is_leaf = graphene.Boolean()
    titos = graphene.String()

    def resolve_frame_id(self, info: ResolveInfo) -> DBID:
        # pyre-fixme[16]: `TraceFrameQueryResultType` has no attribute `id`.
        return self.id

    def resolve_is_leaf(self, info: ResolveInfo) -> bool:
        return self.callee_port in LEAF_NAMES


class TraceFrameQueryResult(NamedTuple):
    id: DBID
    caller: str
    caller_port: str
    callee: str
    callee_port: str
    caller_id: Optional[DBID] = None
    callee_id: Optional[DBID] = None
    callee_location: Optional[SourceLocation] = None
    kind: Optional[TraceKind] = None
    filename: Optional[str] = None
    trace_length: Optional[int] = None
    titos: Optional[str] = None
    type_interval_lower: Optional[int] = None
    type_interval_upper: Optional[int] = None
    preserves_type_context: Optional[bool] = None
    shared_texts: List[SharedText] = []

    @staticmethod
    def from_record(
        record: Any,  # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        shared_texts: Optional[List[SharedText]] = None,
    ) -> "TraceFrameQueryResult":
        return TraceFrameQueryResult(
            id=record.id,
            caller=record.caller,
            caller_port=record.caller_port,
            callee=record.callee,
            callee_port=record.callee_port,
            caller_id=record.caller_id,
            callee_id=record.callee_id,
            callee_location=record.callee_location,
            kind=record.kind,
            filename=record.filename,
            trace_length=getattr(record, "trace_length", None),
            titos=";".join([str(titos) for titos in getattr(record, "titos", [])]),
            type_interval_lower=record.type_interval_lower,
            type_interval_upper=record.type_interval_upper,
            preserves_type_context=record.preserves_type_context,
            shared_texts=shared_texts if shared_texts else [],
        )

    def is_leaf(self) -> bool:
        return self.callee_port in LEAF_NAMES


class TraceTuple(NamedTuple):
    trace_frame: TraceFrameQueryResult
    branches: int = 1
    missing: bool = False
    # Suppose we select a trace frame (A->B) and the generated trace is
    #   (A->B), (B->C), (C->D) with D as leaf.
    # When we display traces, we only use the callee, so this trace would look
    #   like B->C->D. If we also want to see A->, then we need to add a
    #   placeholder trace tuple. We do this by setting our trace tuples to
    #   [(A->B, placeholder=True), (A->B), (B->C), (C->D)]. When placeholder is
    #   True, that means we need to output the caller rather than the callee.
    placeholder: bool = False


class LeafLookup:
    def __init__(
        self, sources: Dict[int, str], sinks: Dict[int, str], features: Dict[int, str]
    ) -> None:
        self._lookup: Dict[SharedTextKind, Dict[int, str]] = {
            SharedTextKind.SOURCE: sources,
            SharedTextKind.SINK: sinks,
            SharedTextKind.FEATURE: features,
        }

    @staticmethod
    def create(session: Session) -> "LeafLookup":
        return LeafLookup(
            {
                int(id): contents
                for id, contents in session.query(
                    SharedText.id, SharedText.contents
                ).filter(SharedText.kind == SharedTextKind.SOURCE)
            },
            {
                int(id): contents
                for id, contents in session.query(
                    SharedText.id, SharedText.contents
                ).filter(SharedText.kind == SharedTextKind.SINK)
            },
            {
                int(id): contents
                for id, contents in session.query(
                    SharedText.id, SharedText.contents
                ).filter(SharedText.kind == SharedTextKind.FEATURE)
            },
        )

    def resolve(self, ids: Sequence[int], kind: SharedTextKind) -> Set[str]:
        if kind not in [
            SharedTextKind.SOURCE,
            SharedTextKind.SINK,
            SharedTextKind.FEATURE,
        ]:
            raise ValueError(f"Cannot resolve ids of kind `{kind}`")

        lookup = self._lookup[kind]
        return {lookup[id] for id in ids if id in lookup}


def initial_frames(
    session: Session,
    issue_id: DBID,
    kind: TraceKind,
) -> List[TraceFrameQueryResult]:

    records = list(
        session.query(
            TraceFrame.id,
            TraceFrame.caller_id,
            CallerText.contents.label("caller"),
            TraceFrame.caller_port,
            TraceFrame.callee_id,
            CalleeText.contents.label("callee"),
            TraceFrame.callee_port,
            TraceFrame.callee_location,
            TraceFrame.kind,
            TraceFrame.type_interval_lower,
            TraceFrame.type_interval_upper,
            TraceFrame.preserves_type_context,
            FilenameText.contents.label("filename"),
            TraceFrameLeafAssoc.trace_length,
            TraceFrame.titos,
        )
        .filter(TraceFrame.kind == kind)
        .join(
            IssueInstanceTraceFrameAssoc,
            IssueInstanceTraceFrameAssoc.trace_frame_id == TraceFrame.id,
        )
        .filter(IssueInstanceTraceFrameAssoc.issue_instance_id == issue_id)
        .join(CallerText, CallerText.id == TraceFrame.caller_id)
        .join(CalleeText, CalleeText.id == TraceFrame.callee_id)
        .join(FilenameText, FilenameText.id == TraceFrame.filename_id)
        .join(TraceFrameLeafAssoc, TraceFrameLeafAssoc.trace_frame_id == TraceFrame.id)
        .group_by(TraceFrame.id)
        .order_by(TraceFrameLeafAssoc.trace_length, TraceFrame.callee_location)
        .all()
    )

    frames = []
    for record in records:
        shared_texts = list(
            session.query(SharedText)
            .join(TraceFrameLeafAssoc, SharedText.id == TraceFrameLeafAssoc.leaf_id)
            .filter(TraceFrameLeafAssoc.trace_frame_id == record.id)
            .all()
        )
        frames.append(TraceFrameQueryResult.from_record(record, shared_texts))
    return frames


def navigate_trace_frames(
    session: Session,
    initial_trace_frames: List[TraceFrameQueryResult],
    sources: Set[str],
    sinks: Set[str],
    index: int = 0,
) -> List[Tuple[TraceFrameQueryResult, int]]:
    leaf_lookup = LeafLookup.create(session)

    if not initial_trace_frames:
        return []
    trace_frames = [(initial_trace_frames[index], len(initial_trace_frames))]
    visited_ids: Set[int] = {int(initial_trace_frames[index].id)}
    while not trace_frames[-1][0].is_leaf():
        trace_frame, branches = trace_frames[-1]
        if trace_frame.kind == TraceKind.POSTCONDITION:
            leaf_kind = sources
        elif trace_frame.kind == TraceKind.PRECONDITION:
            leaf_kind = sinks
        else:
            assert (
                trace_frame.kind == TraceKind.POSTCONDITION
                or trace_frame.kind == TraceKind.PRECONDITION
            )
        next_nodes = next_frames(
            session,
            trace_frame,
            # pyre-fixme[61]: `leaf_kind` may not be initialized here.
            leaf_kind,
            visited_ids,
            leaf_lookup=leaf_lookup,
        )

        if len(next_nodes) == 0:
            # Denote a missing frame by setting caller to None
            trace_frames.append(
                (
                    TraceFrameQueryResult(
                        id=DBID(0),
                        callee=trace_frame.callee,
                        callee_port=trace_frame.callee_port,
                        caller="",
                        caller_port="",
                    ),
                    0,
                )
            )
            return trace_frames

        visited_ids.add(int(next_nodes[0].id))
        trace_frames.append((next_nodes[0], len(next_nodes)))
    return trace_frames


def next_frames(
    session: Session,
    pre_frame: TraceFrameQueryResult,
    leaf_kinds: Set[str],
    visited_ids: Set[int],
    run_id: Optional[DBID] = None,
    leaf_lookup: Optional[LeafLookup] = None,
    backwards: bool = False,
) -> List[TraceFrameQueryResult]:
    """Finds all trace frames that the given trace_frame flows to.

    When backwards=True, the result will include the parameter trace_frame,
    since we are filtering on the parameter's callee.
    """
    query = (
        session.query(
            TraceFrame.id,
            TraceFrame.caller_id,
            CallerText.contents.label("caller"),
            TraceFrame.caller_port,
            TraceFrame.callee_id,
            CalleeText.contents.label("callee"),
            TraceFrame.callee_port,
            TraceFrame.callee_location,
            TraceFrame.kind,
            TraceFrame.type_interval_lower,
            TraceFrame.type_interval_upper,
            TraceFrame.preserves_type_context,
            FilenameText.contents.label("filename"),
            TraceFrameLeafAssoc.trace_length,
            TraceFrame.titos,
        )
        .filter(TraceFrame.run_id == (run_id or run.latest(session)))
        .filter(TraceFrame.kind == pre_frame.kind)
        .join(CallerText, CallerText.id == TraceFrame.caller_id)
        .join(CalleeText, CalleeText.id == TraceFrame.callee_id)
        .join(FilenameText, FilenameText.id == TraceFrame.filename_id)
        .filter(
            TraceFrame.caller_id != TraceFrame.callee_id
        )  # skip recursive calls for now
    )
    if backwards:
        query = query.filter(TraceFrame.callee_id == pre_frame.caller_id).filter(
            TraceFrame.callee_port == pre_frame.caller_port
        )
    else:
        query = query.filter(TraceFrame.caller_id == pre_frame.callee_id).filter(
            TraceFrame.caller_port == pre_frame.callee_port
        )

    results = (
        query.join(
            TraceFrameLeafAssoc, TraceFrameLeafAssoc.trace_frame_id == TraceFrame.id
        )
        .group_by(TraceFrame.id)
        .order_by(TraceFrameLeafAssoc.trace_length, TraceFrame.callee_location)
    )

    filtered_results = []

    for frame in results:
        if int(frame.id) in visited_ids:
            continue
        if not leaf_kinds.intersection(
            set(
                get_leaves_trace_frame(
                    session,
                    int(frame.id),
                    trace_kind_to_shared_text_kind(frame.kind),
                    _leaf_lookup(session, leaf_lookup),
                )
            )
        ):
            continue

        if not TraceFrame.type_intervals_match_or_ignored(
            pre_frame.type_interval_lower,
            pre_frame.type_interval_upper,
            pre_frame.preserves_type_context,
            frame.type_interval_lower,
            frame.type_interval_upper,
            frame.preserves_type_context,
        ):
            # We are not ignoring and we have no match
            # so therefore filter the result from the results
            continue

        shared_texts = list(
            session.query(SharedText)
            .join(TraceFrameLeafAssoc, SharedText.id == TraceFrameLeafAssoc.leaf_id)
            .filter(TraceFrameLeafAssoc.trace_frame_id == frame.id)
            .all()
        )

        filtered_results.append((frame, shared_texts))

    return [
        TraceFrameQueryResult.from_record(frame, shared_texts)
        for frame, shared_texts in filtered_results
    ]


def _leaf_lookup(
    session: Session, leaf_lookup: Optional[LeafLookup] = None
) -> LeafLookup:
    if leaf_lookup:
        return leaf_lookup

    return LeafLookup.create(session)


def get_leaves_trace_frame(
    session: Session,
    trace_frame_id: Union[int, DBID],
    kind: SharedTextKind,
    leaf_lookup: Optional[LeafLookup] = None,
) -> Set[str]:
    ids = [
        int(id)
        for id, in session.query(SharedText.id)
        .distinct(SharedText.id)
        .join(TraceFrameLeafAssoc, SharedText.id == TraceFrameLeafAssoc.leaf_id)
        .filter(TraceFrameLeafAssoc.trace_frame_id == trace_frame_id)
        .filter(SharedText.kind == kind)
    ]
    return _leaf_lookup(session, leaf_lookup).resolve(ids, kind)


def trace_kind_to_shared_text_kind(trace_kind: Optional[TraceKind]) -> SharedTextKind:
    if trace_kind == TraceKind.POSTCONDITION:
        return SharedTextKind.SOURCE
    if trace_kind == TraceKind.PRECONDITION:
        return SharedTextKind.SINK

    raise AssertionError(f"{trace_kind} is invalid")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import List, Optional, Set, Tuple

from . import DictEntries, PipelineStep, Summary

log: logging.Logger = logging.getLogger("sapp")


class AddFeatures(PipelineStep[DictEntries, DictEntries]):
    """Optional pipeline that attaches given features to all issues.
    If the features list is empty, the input is simply passed without
    change as the output"""

    def __init__(self, features: Optional[List[str]]) -> None:
        super().__init__()
        self.features: Set[str] = set(features or [])

    def run(self, input: DictEntries, summary: Summary) -> Tuple[DictEntries, Summary]:
        if len(self.features) > 0:
            log.info("Attaching provided features")
            input["issues"] = [
                issue._replace(features=list(set(issue.features) | self.features))
                for issue in input["issues"]
            ]
        return input, summary



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

#!/usr/bin/env python3

import collections
import logging
from typing import Optional, Tuple

from ..bulk_saver import BulkSaver
from ..db import DB
from ..decorators import log_time
from ..models import (
    Issue,
    MetaRunToRunAssoc,
    PrimaryKeyGenerator,
    Run,
    RunStatus,
    RunSummary,
    TraceFrame,
    TraceFrameAnnotation,
    TraceFrameLeafAssoc,
    TraceKind,
)
from ..trace_graph import TraceGraph
from . import PipelineStep, Summary

# pyre-fixme[5]: Global expression must be annotated.
log = logging.getLogger("sapp")


# pyre-fixme[13]: Attribute `summary` is never initialized.
class DatabaseSaver(PipelineStep[TraceGraph, RunSummary]):
    RUN_MODEL = Run

    # pyre-fixme[3]: Return type must be annotated.
    def __init__(
        self,
        database: DB,
        primary_key_generator: Optional[PrimaryKeyGenerator] = None,
        dry_run: bool = False,
    ):
        # pyre-fixme[4]: Attribute must be annotated.
        self.dbname = database.dbname
        self.database = database
        # pyre-fixme[4]: Attribute must be annotated.
        self.primary_key_generator = primary_key_generator or PrimaryKeyGenerator()
        self.bulk_saver = BulkSaver(self.primary_key_generator)
        self.dry_run = dry_run
        self.graph: TraceGraph
        self.summary: Summary

    @log_time
    def run(self, input: TraceGraph, summary: Summary) -> Tuple[RunSummary, Summary]:
        self.graph = input
        self.summary = summary

        self._prep_save()
        return self._save(), self.summary

    # pyre-fixme[3]: Return type must be annotated.
    def _prep_save(self):
        """Prepares the bulk saver to load the trace graph info into the
        database.
        """
        log.info("Preparing bulk save.")
        self.graph.update_bulk_saver(self.bulk_saver)

        for trace_kind, unused in self.summary["trace_entries"].items():
            log.info(
                "Dropped %d unused %s, %d are missing",
                sum(len(v) for v in unused.values()),
                trace_kind,
                len(self.summary["missing_traces"][trace_kind]),
            )

    def _save(self) -> RunSummary:
        """Saves bulk saver's info into the databases in bulk."""
        assert self.summary["run"] is not None, "Must have called process before"

        trace_frames = self.bulk_saver.get_items_to_add(TraceFrame)
        log.info(
            "Saving %d issues, %d trace frames, %d trace annotations, %d trace frame leaf assocs",
            len(self.bulk_saver.get_items_to_add(Issue)),
            len(self.bulk_saver.get_items_to_add(TraceFrame)),
            len(self.bulk_saver.get_items_to_add(TraceFrameAnnotation)),
            len(self.bulk_saver.get_items_to_add(TraceFrameLeafAssoc)),
        )

        num_pre = 0
        num_post = 0
        for frame in trace_frames:
            if frame.kind == TraceKind.PRECONDITION:
                num_pre += 1
            elif frame.kind == TraceKind.POSTCONDITION:
                num_post += 1
        log.info(
            "Within trace frames: %d preconditions, %d postconditions",
            num_pre,
            num_post,
        )

        if not self.dry_run:
            with self.database.make_session() as session:
                pk_gen = self.primary_key_generator.reserve(session, [Run])
                self.summary["run"].id.resolve(id=pk_gen.get(Run), is_new=True)
                session.add(self.summary["run"])
                meta_run_identifier = self.summary.get("meta_run_identifier")
                if meta_run_identifier is not None:
                    session.add(
                        MetaRunToRunAssoc(
                            meta_run_id=meta_run_identifier,
                            run_id=self.summary["run"].id,
                            run_label=self.summary.get("meta_run_child_label", None),
                        )
                    )
                session.commit()

                run_id = self.summary["run"].id.resolved()
                self.summary["run"] = None  # Invalidate it

            self.bulk_saver.save_all(self.database)

            # Now that the run is finished, fetch it from the DB again and set its
            # status to FINISHED.
            with self.database.make_session() as session:
                run = session.query(self.RUN_MODEL).filter_by(id=run_id).one()
                run.status = RunStatus.FINISHED
                session.add(run)
                session.commit()
                run_summary = run.get_summary()
        else:
            run_summary = self._get_dry_run_summary()

        # pyre-fixme[16]: `RunSummary` has no attribute `num_invisible_issues`.
        run_summary.num_invisible_issues = 0
        run_summary.num_missing_preconditions = len(
            self.summary["missing_traces"][TraceKind.precondition]
        )
        run_summary.num_missing_postconditions = len(
            self.summary["missing_traces"][TraceKind.postcondition]
        )

        return run_summary

    def _get_dry_run_summary(self) -> RunSummary:
        run = self.summary["run"]
        return RunSummary(
            commit_hash=run.commit_hash,
            differential_id=run.differential_id,
            id=None,
            job_id=run.job_id,
            num_new_issues=0,
            num_total_issues=self.graph.get_number_issues(),
            alarm_counts=dict(
                collections.Counter(issue.code for issue in self.graph.get_issues())
            ),
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from pathlib import Path
from typing import Any, List, Optional

import graphene
from graphene import relay
from graphene_sqlalchemy import get_session
from graphql.execution.base import ResolveInfo

from ..filter import Filter
from ..models import DBID, TraceFrame, TraceKind
from . import filters as filters_module
from . import issues, run, trace, typeahead
from .issues import (
    Instance,
    IssueQueryResult,
    IssueQueryResultType,
    update_status,
)
from .trace import TraceFrameQueryResult, TraceFrameQueryResultType


class RunConnection(relay.Connection):
    class Meta:
        node = run.Run


class IssueConnection(relay.Connection):
    class Meta:
        node = IssueQueryResultType


class TraceFrameConnection(relay.Connection):
    class Meta:
        node = TraceFrameQueryResultType


class File(graphene.ObjectType):
    class Meta:
        interfaces = (relay.Node,)

    path = graphene.String()
    contents = graphene.String()
    editor_link = graphene.String()


class FileConnection(relay.Connection):
    class Meta:
        node = File


class CodeConnection(relay.Connection):
    class Meta:
        node = typeahead.Code


class PathConnection(relay.Connection):
    class Meta:
        node = typeahead.Path


class CallableConnection(relay.Connection):
    class Meta:
        node = typeahead.Callable


class SourceNameConnection(relay.Connection):
    class Meta:
        node = typeahead.SourceName


class SourceKindConnection(relay.Connection):
    class Meta:
        node = typeahead.SourceKind


class SinkNameConnection(relay.Connection):
    class Meta:
        node = typeahead.SinkName


class SinkKindConnection(relay.Connection):
    class Meta:
        node = typeahead.SinkKind


class StatusConnection(relay.Connection):
    class Meta:
        node = typeahead.Status


class FeatureConnection(relay.Connection):
    class Meta:
        node = typeahead.Feature


class FilterConnection(relay.Connection):
    class Meta:
        node = filters_module.Filter


class FeatureCondition(graphene.InputObjectType):
    mode = graphene.String()
    features = graphene.List(graphene.String)


class MatchesIsField(graphene.InputObjectType):
    operation = graphene.String()
    value = graphene.List(graphene.String)


class Query(graphene.ObjectType):
    # pyre-fixme[4]: Attribute must be annotated.
    node = relay.Node.Field()

    runs = relay.ConnectionField(
        RunConnection,
    )

    issues = relay.ConnectionField(
        IssueConnection,
        run_id=graphene.Int(required=True),
        codes=graphene.List(graphene.Int, default_value=["%"]),
        paths=graphene.List(graphene.String, default_value=["%"]),
        callables=MatchesIsField(),
        source_names=MatchesIsField(),
        source_kinds=MatchesIsField(),
        sink_names=MatchesIsField(),
        sink_kinds=MatchesIsField(),
        statuses=graphene.List(graphene.String, default_value=["%"]),
        features=graphene.List(FeatureCondition),
        min_trace_length_to_sinks=graphene.Int(),
        max_trace_length_to_sinks=graphene.Int(),
        min_trace_length_to_sources=graphene.Int(),
        max_trace_length_to_sources=graphene.Int(),
        issue_instance_id=graphene.Int(),
        is_new_issue=graphene.Boolean(),
    )

    trace = relay.ConnectionField(TraceFrameConnection, issue_instance_id=graphene.ID())
    initial_trace_frames = relay.ConnectionField(
        TraceFrameConnection, issue_instance_id=graphene.Int(), kind=graphene.String()
    )
    next_trace_frames = relay.ConnectionField(
        TraceFrameConnection,
        issue_instance_id=graphene.Int(),
        frame_id=graphene.Int(),
        kind=graphene.String(),
    )

    # Typeahead data.
    codes = relay.ConnectionField(CodeConnection)
    paths = relay.ConnectionField(PathConnection)
    callables = relay.ConnectionField(CallableConnection)
    features = relay.ConnectionField(FeatureConnection)
    source_names = relay.ConnectionField(SourceNameConnection)
    source_kinds = relay.ConnectionField(SourceKindConnection)
    sink_names = relay.ConnectionField(SinkNameConnection)
    sink_kinds = relay.ConnectionField(SinkKindConnection)
    statuses = relay.ConnectionField(StatusConnection)

    file = relay.ConnectionField(FileConnection, path=graphene.String())

    filters = relay.ConnectionField(FilterConnection)

    def resolve_runs(self, info: ResolveInfo) -> List[run.Run]:
        session = get_session(info.context)
        return run.runs(session)

    def resolve_issues(
        self,
        info: ResolveInfo,
        run_id: int,
        codes: List[int],
        paths: List[str],
        statuses: List[str],
        callables: Optional[MatchesIsField] = None,
        features: Optional[List[FeatureCondition]] = None,
        min_trace_length_to_sinks: Optional[int] = None,
        max_trace_length_to_sinks: Optional[int] = None,
        min_trace_length_to_sources: Optional[int] = None,
        max_trace_length_to_sources: Optional[int] = None,
        issue_instance_id: Optional[int] = None,
        is_new_issue: Optional[bool] = None,
        source_names: Optional[MatchesIsField] = None,
        source_kinds: Optional[MatchesIsField] = None,
        sink_names: Optional[MatchesIsField] = None,
        sink_kinds: Optional[MatchesIsField] = None,
        **kwargs: Any,
    ) -> List[IssueQueryResult]:
        session = get_session(info.context)

        filter_instance = Filter.from_query(
            codes,
            paths,
            callables,
            statuses,
            source_names,
            source_kinds,
            sink_names,
            sink_kinds,
            features,
            min_trace_length_to_sinks,
            max_trace_length_to_sinks,
            min_trace_length_to_sources,
            max_trace_length_to_sources,
            is_new_issue,
        )

        issues = (
            Instance(session, DBID(run_id))
            .where_filter(filter_instance)
            .where_issue_instance_id_is(issue_instance_id)
            .get()
        )
        return issues

    def resolve_initial_trace_frames(
        self, info: ResolveInfo, issue_instance_id: int, kind: str
    ) -> List[TraceFrameQueryResult]:
        session = info.context.get("session")
        return trace.initial_frames(
            session, DBID(issue_instance_id), TraceKind.create_from_string(kind)
        )

    def resolve_next_trace_frames(
        self, info: ResolveInfo, issue_instance_id: int, frame_id: int, kind: str
    ) -> List[TraceFrameQueryResult]:
        session = info.context.get("session")

        trace_kind = TraceKind.create_from_string(kind)
        if trace_kind == TraceKind.POSTCONDITION:
            leaf_kind = issues.sources(session, DBID(issue_instance_id))
        elif trace_kind == TraceKind.PRECONDITION:
            leaf_kind = issues.sinks(session, DBID(issue_instance_id))

        trace_frame = session.query(TraceFrame).get(frame_id)
        if trace_frame is None:
            raise ValueError(f"`{frame_id}` is not a valid trace frame id")

        return trace.next_frames(
            session,
            trace_frame,
            # pyre-fixme[61]: `leaf_kind` may not be initialized here.
            leaf_kind,
            visited_ids=set(),
        )

    def resolve_codes(self, info: ResolveInfo) -> List[typeahead.Code]:
        session = info.context["session"]
        return typeahead.all_codes(session)

    def resolve_paths(self, info: ResolveInfo) -> List[typeahead.Path]:
        session = info.context["session"]
        return typeahead.all_paths(session)

    def resolve_callables(self, info: ResolveInfo) -> List[typeahead.Callable]:
        session = info.context["session"]
        return typeahead.all_callables(session)

    def resolve_source_names(self, info: ResolveInfo) -> List[typeahead.SourceName]:
        session = info.context["session"]
        return typeahead.all_source_names(session)

    def resolve_source_kinds(self, info: ResolveInfo) -> List[typeahead.SourceName]:
        session = info.context["session"]
        return typeahead.all_source_kinds(session)

    def resolve_sink_names(self, info: ResolveInfo) -> List[typeahead.SourceName]:
        session = info.context["session"]
        return typeahead.all_sink_names(session)

    def resolve_sink_kinds(self, info: ResolveInfo) -> List[typeahead.SourceName]:
        session = info.context["session"]
        return typeahead.all_sink_kinds(session)

    def resolve_statuses(self, info: ResolveInfo) -> List[typeahead.Status]:
        session = info.context["session"]
        return typeahead.all_statuses(session)

    def resolve_features(self, info: ResolveInfo) -> List[typeahead.Feature]:
        session = info.context["session"]
        return typeahead.all_features(session)

    def resolve_file(self, info: ResolveInfo, path: str, **kwargs: Any) -> List[File]:
        if ".." in path:
            raise FileNotFoundError("Attempted directory traversal")

        source_directory = Path(info.context.get("source_directory") or os.getcwd())
        full_path = source_directory / path
        contents = full_path.read_text()

        editor_link = None
        editor_schema = info.context.get("editor_schema")
        if editor_schema is not None:
            editor_link = f"{editor_schema}//file/{str(full_path)}"

        return [File(path=path, contents=contents, editor_link=editor_link)]

    def resolve_filters(self, info: ResolveInfo) -> List[filters_module.Filter]:
        session = info.context["session"]
        return filters_module.all_filters(session)


class SaveFilterMutation(relay.ClientIDMutation):
    node = graphene.Field(filters_module.Filter)

    class Input:
        name = graphene.String(required=True)
        description = graphene.String()
        json = graphene.String()

    def mutate_and_get_payload(
        self, info: ResolveInfo, **kwargs: Any
    ) -> "SaveFilterMutation":
        session = info.context.get("session")
        filter = filters_module.Filter(**kwargs)
        filters_module.save_filter(session, filter)
        return SaveFilterMutation(node=filter)


class DeleteFilterMutation(relay.ClientIDMutation):
    class Input:
        name = graphene.String(required=True)

    def mutate_and_get_payload(
        self, info: ResolveInfo, **kwargs: Any
    ) -> "DeleteFilterMutation":
        session = info.context.get("session")
        filters_module.delete_filter(session, kwargs["name"])
        return DeleteFilterMutation()


class DeleteRunMutation(relay.ClientIDMutation):
    class Input:
        id = graphene.ID(required=True)

    def mutate_and_get_payload(
        self, info: ResolveInfo, **kwargs: Any
    ) -> "DeleteRunMutation":
        session = info.context.get("session")
        run.delete_run(session, kwargs["id"])
        return DeleteRunMutation()


class UpdateStatusMutation(relay.ClientIDMutation):
    class Input:
        id = graphene.ID(required=True)
        status = graphene.String(required=True)

    def mutate_and_get_payload(
        self, info: ResolveInfo, **kwargs: Any
    ) -> "UpdateStatusMutation":
        session = info.context.get("session")
        update_status(session, kwargs["id"], kwargs["status"])
        return UpdateStatusMutation()


class Mutation(graphene.ObjectType):
    # pyre-fixme[4]: Attribute must be annotated.
    save_filter = SaveFilterMutation.Field()
    # pyre-fixme[4]: Attribute must be annotated.
    delete_filter = DeleteFilterMutation.Field()
    delete_run: DeleteRunMutation = DeleteRunMutation.Field()
    update_status: UpdateStatusMutation = UpdateStatusMutation.Field()


schema = graphene.Schema(query=Query, mutation=Mutation, auto_camelcase=False)
